# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from pathlib import Path
from typing import Optional

from openc3.environment import OPENC3_SCOPE
from openc3.models.model import Model
from openc3.models.secret_model import SecretModel
from openc3.models.target_model import TargetModel
from openc3.models.microservice_model import MicroserviceModel
from openc3.logs.stream_log_pair import StreamLogPair
from openc3.top_level import get_class_from_module
from openc3.utilities.string import filename_to_module, filename_to_class_name


class InterfaceModel(Model):
    INTERFACES_PRIMARY_KEY = "openc3_interfaces"
    ROUTERS_PRIMARY_KEY = "openc3_routers"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(f"{scope}__{cls._get_key()}", name)

    @classmethod
    def names(cls, scope: str):
        return super().names(f"{scope}__{cls._get_key()}")

    @classmethod
    def all(cls, scope: str):
        return super().all(f"{scope}__{cls._get_key()}")

    # END NOTE

    @classmethod
    def _get_type(cls):
        """Helper method to return the correct type based on class name"""
        return cls.__name__.split("Model")[0].upper()

    @classmethod
    def _get_key(cls):
        """Helper method to return the correct primary key based on class name"""
        type = cls._get_type()
        match type:
            case "INTERFACE":
                return cls.INTERFACES_PRIMARY_KEY
            case "ROUTER":
                return cls.ROUTERS_PRIMARY_KEY
            case _:
                raise RuntimeError(f"Unknown type {type} from class {cls.__name__}")

    def __init__(
        self,
        name,
        config_params: Optional[list] = None,
        target_names: Optional[list] = None,
        cmd_target_names: Optional[list] = None,
        tlm_target_names: Optional[list] = None,
        connect_on_startup: bool = True,
        auto_reconnect: bool = True,
        reconnect_delay: float = 5.0,
        disable_disconnect: bool = False,
        options: Optional[list] = None,
        secret_options: Optional[list] = None,
        protocols: Optional[list] = None,
        log_stream=None,
        updated_at: Optional[float] = None,
        plugin: Optional[str] = None,
        needs_dependencies: bool = False,
        secrets: Optional[list] = None,
        cmd: str = None,
        work_dir: str = "/openc3/lib/openc3/microservices",
        ports: Optional[list] = None,
        env: Optional[dict] = None,
        container=None,
        prefix=None,
        shard=0,
        scope: str = OPENC3_SCOPE,
    ):
        type = self.__class__._get_type()
        if type == "INTERFACE":
            super().__init__(
                f"{scope}__{InterfaceModel.INTERFACES_PRIMARY_KEY}",
                name=name,
                updated_at=updated_at,
                plugin=plugin,
                scope=scope,
            )
        else:
            super().__init__(
                f"{scope}__{InterfaceModel.ROUTERS_PRIMARY_KEY}",
                name=name,
                updated_at=updated_at,
                plugin=plugin,
                scope=scope,
            )
        self.config_params = [] if config_params is None else config_params
        self.target_names = [] if target_names is None else target_names
        self.cmd_target_names = [] if cmd_target_names is None else cmd_target_names
        self.tlm_target_names = [] if tlm_target_names is None else tlm_target_names
        self.connect_on_startup = connect_on_startup
        self.auto_reconnect = auto_reconnect
        self.reconnect_delay = reconnect_delay
        self.disable_disconnect = disable_disconnect
        self.options = [] if options is None else options
        self.secret_options = [] if secret_options is None else secret_options
        self.protocols = [] if protocols is None else protocols
        self.log_stream = log_stream
        self.needs_dependencies = needs_dependencies
        self.cmd = cmd
        if self.cmd is None:
            microservice_name = f"{self.scope}__{type}__{self.name}"
            if len(self.config_params) == 0 or Path(config_params[0]).suffix == ".py":
                work_dir = work_dir.replace("openc3/lib", "openc3/python")
                self.cmd = [
                    "python",
                    f"{type.lower()}_microservice.py",
                    microservice_name,
                ]
            else:
                raise RuntimeError(f"Unknown file type {config_params[0]}")
        self.work_dir = work_dir
        self.ports = [] if ports is None else ports
        self.env = {} if env is None else env
        self.container = container
        self.prefix = prefix
        self.shard = shard
        if self.shard is None:
            self.shard = 0
        self.secrets = [] if secrets is None else secrets

    # Called by InterfaceMicroservice to instantiate the Interface defined
    # by the model configuration. Must be called after get_model which
    # calls from_json to instantiate the class and populate the attributes.
    def build(self):
        klass = get_class_from_module(
            filename_to_module(self.config_params[0]),
            filename_to_class_name(self.config_params[0]),
        )
        if len(self.config_params) > 1:
            interface_or_router = klass(*self.config_params[1:])
        else:
            interface_or_router = klass()
        interface_or_router.secrets.setup(self.secrets)
        interface_or_router.target_names = self.target_names[:]
        interface_or_router.cmd_target_names = self.cmd_target_names[:]
        interface_or_router.tlm_target_names = self.tlm_target_names[:]
        interface_or_router.connect_on_startup = self.connect_on_startup
        interface_or_router.auto_reconnect = self.auto_reconnect
        interface_or_router.reconnect_delay = self.reconnect_delay
        interface_or_router.disable_disconnect = self.disable_disconnect
        for option in self.options:
            interface_or_router.set_option(option[0], option[1:])
        for option in self.secret_options:
            secret_name = option[1]
            secret_value = interface_or_router.secrets.get(secret_name, scope=self.scope)
            interface_or_router.set_option(option[0], [secret_value])
        for protocol in self.protocols:
            klass = get_class_from_module(
                filename_to_module(protocol[1]),
                filename_to_class_name(protocol[1]),
            )
            interface_or_router.add_protocol(klass, protocol[2:], protocol[0].upper())
        if self.log_stream:
            interface_or_router.stream_log_pair = StreamLogPair(interface_or_router.name, self.log_stream)
            interface_or_router.start_raw_logging
        return interface_or_router

    def as_json(self):
        return {
            "name": self.name,
            "config_params": self.config_params,
            "target_names": self.target_names,
            "cmd_target_names": self.cmd_target_names,
            "tlm_target_names": self.tlm_target_names,
            "connect_on_startup": self.connect_on_startup,
            "auto_reconnect": self.auto_reconnect,
            "reconnect_delay": self.reconnect_delay,
            "disable_disconnect": self.disable_disconnect,
            "options": self.options,
            "secret_options": self.secret_options,
            "protocols": self.protocols,
            "log_stream": self.log_stream,
            "plugin": self.plugin,
            "needs_dependencies": self.needs_dependencies,
            "secrets": (self.secrets.as_json() if isinstance(self.secrets, SecretModel) else self.secrets),
            "cmd": self.cmd,
            "work_dir": self.work_dir,
            "ports": self.ports,
            "env": self.env,
            "container": self.container,
            "prefix": self.prefix,
            "shard": self.shard,
            "updated_at": self.updated_at,
        }

    def ensure_target_exists(self, target_name):
        target = TargetModel.get(name=target_name, scope=self.scope)
        if not target:
            raise RuntimeError(f"Target {target_name} does not exist")
        return target

    def unmap_target(self, target_name, cmd_only=False, tlm_only=False):
        if cmd_only and tlm_only:
            cmd_only = False
            tlm_only = False
        target_name = str(target_name).upper()

        # Remove from this interface
        if cmd_only:
            self.cmd_target_names.remove(target_name)
            if target_name not in self.tlm_target_names:
                self.target_names.remove(target_name)
        elif tlm_only:
            self.tlm_target_names.remove(target_name)
            if target_name not in self.cmd_target_names:
                self.target_names.remove(target_name)
        else:
            self.cmd_target_names.remove(target_name)
            self.tlm_target_names.remove(target_name)
            self.target_names.remove(target_name)
        self.update()

        # Respawn the microservice
        type = self.__class__.__name__.split("Model")[0].upper()
        microservice_name = f"{self.scope}__{type}__{self.name}"
        microservice = MicroserviceModel.get_model(name=microservice_name, scope=self.scope)
        if target_name not in self.target_names:
            microservice.target_names.remove(target_name)
        microservice.update()

    def map_target(self, target_name, cmd_only=False, tlm_only=False, unmap_old=True):
        if cmd_only and tlm_only:
            cmd_only = False
            tlm_only = False
        target_name = str(target_name).upper()
        self.ensure_target_exists(target_name)

        if unmap_old:
            # Remove from old interface
            all_interfaces = InterfaceModel.all(scope=self.scope)
            old_interface = None
            for _, old_interface_details in all_interfaces.items():
                if target_name in old_interface_details["target_names"]:
                    old_interface = InterfaceModel.from_json(old_interface_details, scope=self.scope)
                    if old_interface:
                        old_interface.unmap_target(target_name, cmd_only=cmd_only, tlm_only=tlm_only)

        # Add to this interface
        if target_name not in self.target_names:
            self.target_names.append(target_name)
        if target_name not in self.cmd_target_names or tlm_only:
            self.cmd_target_names.append(target_name)
        if target_name not in self.tlm_target_names or cmd_only:
            self.tlm_target_names.append(target_name)
        self.update()

        # Respawn the microservice
        type = self.__class__.__name__.split("Model")[0].upper()
        microservice_name = f"{self.scope}__{type}__{self.name}"
        microservice = MicroserviceModel.get_model(name=microservice_name, scope=self.scope)
        if target_name not in microservice.target_names:
            microservice.target_names.append(target_name)
        microservice.update()



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import re
import time
from queue import SimpleQueue, Empty
from openc3.config.config_parser import ConfigParser
from openc3.system.system import System
from openc3.packets.packet import Packet
from openc3.utilities.logger import Logger
from openc3.interfaces.protocols.terminated_protocol import TerminatedProtocol


# Protocol which delineates packets using delimiter characters. Designed for
# text based protocols which expect a command and send a response. The
# protocol handles sending the command and capturing the response.
class TemplateProtocol(TerminatedProtocol):
    # @param write_termination_characters (see TerminatedProtocol#initialize)
    # @param read_termination_characters (see TerminatedProtocol#initialize)
    # @param ignore_lines [Integer] Number of newline terminated reads to
    #   ignore when processing the response
    # @param initial_read_delay [Integer] Initial delay when connecting before
    #   trying to read
    # @param response_lines [Integer] Number of newline terminated lines which
    #   comprise the response
    # @param strip_read_termination (see TerminatedProtocol#initialize)
    # @param discard_leading_bytes (see TerminatedProtocol#initialize)
    # @param sync_pattern (see TerminatedProtocol#initialize)
    # @param fill_fields (see TerminatedProtocol#initialize)
    # @param response_timeout [Float] Number of seconds to wait before timing out
    #   when waiting for a response
    # @param response_polling_period [Float] Number of seconds to wait between polling
    #   for a response
    # @param raise_exceptions [String] Whether to raise exceptions when errors
    #   occur in the protocol like unexpected responses or response timeouts.
    # @param allow_empty_data [true/false/nil] See Protocol#initialize
    def __init__(
        self,
        write_termination_characters,
        read_termination_characters,
        ignore_lines=0,
        initial_read_delay=None,
        response_lines=1,
        strip_read_termination=True,
        discard_leading_bytes=0,
        sync_pattern=None,
        fill_fields=False,
        response_timeout=5.0,
        response_polling_period=0.02,
        raise_exceptions=False,
        allow_empty_data=None,
    ):
        super().__init__(
            write_termination_characters,
            read_termination_characters,
            strip_read_termination,
            discard_leading_bytes,
            sync_pattern,
            fill_fields,
            allow_empty_data,
        )
        self.response_template = None
        self.response_packet = None
        self.response_target_name = None
        self.response_packets = []
        self.write_block_queue = SimpleQueue()
        self.ignore_lines = int(ignore_lines)
        self.response_lines = int(response_lines)
        self.initial_read_delay = ConfigParser.handle_none(initial_read_delay)
        if self.initial_read_delay is not None:
            self.initial_read_delay = float(initial_read_delay)
        self.response_timeout = ConfigParser.handle_none(response_timeout)
        if self.response_timeout is not None:
            self.response_timeout = float(response_timeout)
        if response_polling_period is not None:
            self.response_polling_period = float(response_polling_period)
        else:
            self.response_polling_period = 0.0
        self.connect_complete_time = None
        self.raise_exceptions = ConfigParser.handle_true_false(raise_exceptions)

    def reset(self):
        super().reset()
        self.initial_read_delay_needed = True

    def connect_reset(self):
        super().connect_reset()
        try:
            while self.write_block_queue.qsize != 0:
                self.write_block_queue.get_nowait()
        except Empty:
            pass

        if self.initial_read_delay:
            self.connect_complete_time = time.time() + self.initial_read_delay

    def disconnect_reset(self):
        super().disconnect_reset()
        self.write_block_queue.put(None)  # Unblock the write block queue

    def read_data(self, data, extra=None):
        if len(data) <= 0:
            return super().read_data(data, extra)

        # Drop all data until the initial_read_delay is complete.
        # This gets rid of unused welcome messages,
        # prompts, and other junk on initial connections
        if self.initial_read_delay and self.initial_read_delay_needed and self.connect_complete_time:
            if time.time() < self.connect_complete_time:
                return ("STOP", extra)
            self.initial_read_delay_needed = False
        return super().read_data(data, extra)

    def read_packet(self, packet):
        if self.response_template and self.response_packet:
            # If lines make it this far they are part of a response
            self.response_packets.append(packet)
            if len(self.response_packets) < (self.ignore_lines + self.response_lines):
                return "STOP"

            for _ in range(0, self.ignore_lines):
                self.response_packets.pop(0)
            response_string = b""
            for _ in range(0, self.response_lines):
                response = self.response_packets.pop(0)
                response_string += response.buffer

            # Grab the response packet specified in the command
            result_packet = System.telemetry.packet(self.response_target_name, self.response_packet).clone()
            result_packet.received_time = None
            if result_packet.id_items:
                for item in result_packet.id_items:
                    result_packet.write_item(item, item.id_value, "RAW")

            # Convert the response template into a Regexp
            response_item_names = []
            response_template = self.response_template
            response_template_items = re.findall(r"<.*?>", self.response_template)

            for item in response_template_items:
                response_item_names.append(item[1:-1])
                response_template = response_template.replace(item, r"(.*)")

            # Scan the response for the variables in brackets <VARIABLE>
            # Write the packet value with each of the values received
            response_values = re.findall(response_template, response_string.decode())
            response_values = [x for x in response_values if len(x) > 0]
            if len(response_values) != len(response_item_names):
                interface_name = ""
                if self.interface:
                    interface_name = self.interface.name
                self.handle_error(f"{interface_name}: Unexpected response: {response_string}")
            else:
                try:
                    for i, value in enumerate(response_values):
                        result_packet.write(response_item_names[i], value)
                except (ValueError, RuntimeError) as error:
                    interface_name = ""
                    if self.interface:
                        interface_name = self.interface.name
                    self.handle_error(f"{interface_name}: Could not write value {value} due to {repr(error)}")

            self.response_packets = []

            # Release the write
            if self.response_template and self.response_packet:
                self.write_block_queue.put(None)

            return result_packet
        else:
            return packet

    def write_packet(self, packet):
        # Make sure we are past the initial data dropping period
        if (
            self.initial_read_delay
            and self.initial_read_delay_needed
            and self.connect_complete_time
            and time.time() < self.connect_complete_time
        ):
            delay_needed = self.connect_complete_time - time.time()
            if delay_needed > 0:
                time.sleep(delay_needed)

        # First grab the response template and response packet (if there is one)
        try:
            self.response_template = packet.read("RSP_TEMPLATE").strip()
            self.response_packet = packet.read("RSP_PACKET").strip()
            self.response_target_name = packet.target_name
            # If the template or packet are empty set them to nil. This allows for
            # the user to remove the RSP_TEMPLATE and RSP_PACKET values and avoid
            # any response timeouts
            if len(self.response_template) == 0 or len(self.response_packet) == 0:
                self.response_template = None
                self.response_packet = None
                self.response_target_name = None
        except Exception:
            # If there is no response template we set to nil
            self.response_template = None
            self.response_packet = None
            self.response_target_name = None

        # Grab the command template because that is all we eventually send
        self.template = packet.read("CMD_TEMPLATE")
        # Create a new packet to populate with the template
        raw_packet = Packet(None, None)
        raw_packet.buffer = bytes(self.template, "ascii")
        raw_packet = super().write_packet(raw_packet)
        if isinstance(raw_packet, str):
            return raw_packet

        data = raw_packet.buffer
        # Scan the template for variables in brackets <VARIABLE>
        # Read these values from the packet and substitute them in the template
        # and in the @response_packet name
        for variable in re.findall(r"<(.*?)>", self.template):
            value = packet.read(variable, "RAW")
            data = data.replace(bytes(f"<{variable}>", "ascii"), bytes(f"{value}", "ascii"))
            if self.response_packet:
                self.response_packet = self.response_packet.replace((f"<{variable}>"), f"{value}")
            raw_packet.buffer = data

        return raw_packet

    def post_write_interface(self, packet, data, extra=None):
        if self.response_template and self.response_packet:
            if self.response_timeout:
                response_timeout_time = time.time() + self.response_timeout
            else:
                response_timeout_time = None

            # Block the write until the response is received
            while True:
                try:
                    self.write_block_queue.get_nowait()
                    break
                except Empty:
                    time.sleep(self.response_polling_period)
                    if response_timeout_time is None:
                        continue
                    if response_timeout_time and time.time() < response_timeout_time:
                        continue
                    interface_name = ""
                    if self.interface:
                        interface_name = self.interface.name
                    self.handle_error(f"{interface_name}: Timeout waiting for response")
                    break

            self.response_template = None
            self.response_packet = None
            self.response_target_name = None
            self.response_packets.clear
        return super().post_write_interface(packet, data, extra)

    def handle_error(self, msg):
        Logger.error(msg)
        if self.raise_exceptions:
            raise RuntimeError(msg)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953

import os
import glob
import zipfile
import traceback
from threading import Lock
from openc3.environment import OPENC3_SCOPE, OPENC3_CONFIG_BUCKET
from openc3.top_level import add_to_search_path
from openc3.utilities.bucket import Bucket
from openc3.utilities.logger import Logger
from openc3.utilities.store import Store
from openc3.config.config_parser import ConfigParser
from openc3.packets.packet_config import PacketConfig
from openc3.packets.commands import Commands
from openc3.packets.telemetry import Telemetry
from openc3.packets.limits import Limits
from openc3.system.target import Target


class System:
    # Declare the System class variables ... they are set in __init__
    targets = {}
    packet_config = None
    commands = None
    telemetry = None
    limits = None

    # Variable that holds the singleton instance
    instance_obj = None

    # Mutex used to ensure that only one instance of System is created
    instance_mutex = Lock()

    # Callbacks to call once instance_obj is created
    post_instance_callbacks = []

    @classmethod
    def limits_set(cls, scope=OPENC3_SCOPE):
        """This line is basically the same code as limits_event_topic.py,
        but we can't import it because it imports system.py and that
        creates a circular reference

        Return:
            [Symbol] The current limits_set of the system returned from Redis
        """
        sets = Store.hgetall(f"{scope}__limits_sets")
        try:
            return list(sets.keys())[list(sets.values()).index(b"true")].decode()
        except ValueError:
            return "DEFAULT"

    @classmethod
    def add_post_instance_callback(cls, callback):
        if System.obj_instance:
            callback()
        else:
            cls.post_instance_callbacks << callback

    @classmethod
    def setup_targets(cls, target_names, base_dir, scope=OPENC3_SCOPE):
        if not System.instance_obj:
            os.makedirs(f"{base_dir}/targets", exist_ok=True)
            bucket = Bucket.getClient()
            for target_name in target_names:
                # Retrieve bucket/targets/target_name/<TARGET>_current.zip
                zip_path = f"{base_dir}/targets/{target_name}_current.zip"
                bucket_key = f"{scope}/target_archives/{target_name}/{target_name}_current.zip"
                Logger.info(f"Retrieving {bucket_key} from targets bucket")
                bucket.get_object(bucket=OPENC3_CONFIG_BUCKET, key=bucket_key, path=zip_path)
                with zipfile.ZipFile(zip_path) as zip_file:
                    zip_file.extractall(f"{base_dir}/targets")

                # Now add any modifications in targets_modified/TARGET/cmd_tlm
                # This adds support for remembering dynamically created packets
                # target.txt must be configured to either use all files in cmd_tlm folder (default)
                # or have a predetermined empty file like dynamic_tlm.txt
                bucket_path = f"{scope}/targets_modified/{target_name}/cmd_tlm"
                _, files = bucket.list_files(bucket=OPENC3_CONFIG_BUCKET, path=bucket_path)
                for file in files:
                    bucket_key = os.path.join(bucket_path, file["name"])
                    local_path = f"{base_dir}/targets/{target_name}/cmd_tlm/{file['name']}"
                    bucket.get_object(bucket=OPENC3_CONFIG_BUCKET, key=bucket_key, path=local_path)

            # Build System from targets
            System.instance(target_names, f"{base_dir}/targets")

    @classmethod
    def instance(cls, target_names=None, target_config_dir=None):
        """Get the singleton instance of System

        Args:
            target_names [Array of target_names]
            target_config_dir Directory where target config folders are

        Returns:
            [System] The System singleton
        """
        if System.instance_obj:
            return System.instance_obj
        if not target_names and not target_config_dir:
            raise Exception("System.instance parameters are required on first call")

        with System.instance_mutex:
            if System.instance_obj:
                return System.instance_obj
            System.instance_obj = cls(target_names, target_config_dir)
            for callback in System.post_instance_callbacks:
                callback()
            return System.instance_obj

    # Dynamically add packets to the system instance
    #
    # @param dynamic_packets [Array of packets]
    # @param cmd_or_tlm [Symbol] :COMMAND or :TELEMETRY
    # @param affect_ids [Boolean] Whether to affect packet id lookup or not
    @classmethod
    def dynamic_update(cls, dynamic_packets, cmd_or_tlm="TELEMETRY", affect_ids=False):
        for packet in dynamic_packets:
            if cmd_or_tlm == "TELEMETRY":
                System.instance_obj.telemetry.dynamic_add_packet(packet, affect_ids=affect_ids)
            else:
                System.instance_obj.commands.dynamic_add_packet(packet, affect_ids=affect_ids)

    # Create a new System object.
    #
    # @param target_names [Array of target names]
    # @param target_config_dir Directory where target config folders are
    def __init__(self, target_names, target_config_dir):
        # Find all the base gem lib directories and add them to the search path
        # Ruby handles this because the gem is installed so lib is in the path
        for path in glob.glob("/gems/gems/**/lib"):
            add_to_search_path(path, True)
        if target_config_dir:
            add_to_search_path(target_config_dir, True)
        System.targets = {}
        System.packet_config = PacketConfig()
        System.commands = Commands(System.packet_config, System)
        System.telemetry = Telemetry(System.packet_config, System)
        System.limits = Limits(System.packet_config, System)
        for target_name in target_names:
            self.add_target(target_name, target_config_dir)

    def add_target(self, target_name, target_config_dir):
        parser = ConfigParser()
        folder_name = f"{target_config_dir}/{target_name}"
        if not os.path.exists(folder_name):
            raise parser.error(f"Target folder must exist '{folder_name}'.")

        target = Target(target_name, target_config_dir)
        System.targets[target.name] = target
        errors = []  # Store all errors processing the cmd_tlm files
        try:
            for cmd_tlm_file in target.cmd_tlm_files:
                self.packet_config.process_file(cmd_tlm_file, target.name)
        except Exception as error:
            trace = "".join(traceback.TracebackException.from_exception(error).format())
            errors.append(f"Error processing {target_name}:\n{trace}")
        if len(errors) != 0:
            raise Exception("\n".join(errors))



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.utilities.secrets import Secrets


class KubernetesSecrets(Secrets):
    pass



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
from openc3.config.config_parser import ConfigParser
from openc3.packets.packet_config import PacketConfig
from openc3.tools.table_manager.table import Table
from openc3.tools.table_manager.table_parser import TableParser
from openc3.tools.table_manager.table_item_parser import TableItemParser
from openc3.conversions.generic_conversion import GenericConversion

class TableConfig(PacketConfig):
    """
    Processes the Table Manager configuration files which define tables. Since
    this class inherits from PacketConfig it only needs to implement Table
    Manager specific keywords. All tables are accessed through the table
    and tables methods.
    """

    @classmethod
    def process_file(cls, filename):
        """
        Process a file and create a TableConfig object

        Args:
            filename: The configuration file name

        Returns:
            New TableConfig object
        """
        instance = cls()
        instance._process_file(filename)
        return instance

    def __init__(self):
        """Create the table configuration"""
        super().__init__()
        # Override commands with the Table::TARGET name to store tables
        self.commands[Table.TARGET] = {}
        self.definitions = {}
        self.last_config = []  # Stores array of [filename, contents]

    @property
    def tables(self):
        """
        All tables defined in the configuration file

        Returns:
            Dict of all tables
        """
        # The tests expect tables to contain a direct reference to the command entries
        # with the TEST key (not nested under TABLE)
        return self.commands[Table.TARGET]

    def definition(self, table_name):
        """
        Get the table definition

        Args:
            table_name: Name of the table

        Returns:
            Table definition for the specific table
        """
        return self.definitions[table_name.upper()]

    @property
    def table_names(self):
        """
        Get all the table names

        Returns:
            List of all the table names
        """
        return list(self.commands[Table.TARGET].keys())

    def table(self, table_name):
        """
        Get a specific table

        Args:
            table_name: Name of the table

        Returns:
            The table object
        """
        return self.tables[table_name.upper()]

    def _process_file(self, filename):
        """
        Processes a OpenC3 table configuration file and uses the keywords to build up
        knowledge of the tables.

        Args:
            filename: The name of the configuration file
        """
        # Partial files are included into another file and thus aren't directly processed
        if os.path.basename(filename)[0] == '_':  # Partials start with underscore
            return

        self.filename = filename
        with open(filename, 'r') as file:
            self.last_config = [os.path.basename(filename), file.read()]

        self.converted_type = None
        self.converted_bit_size = None
        self.proc_text = ''
        self.building_generic_conversion = False
        self.defaults = []

        parser = ConfigParser("https://openc3.com/docs/tools/#table-manager-configuration-openc3--39")

        for keyword, params in parser.parse_file(filename):
            if self.building_generic_conversion:
                if keyword in ('GENERIC_READ_CONVERSION_END', 'GENERIC_WRITE_CONVERSION_END'):
                    parser.verify_num_parameters(0, 0, keyword)
                    if 'READ' in keyword:
                        self.current_item.read_conversion = GenericConversion(
                            self.proc_text,
                            self.converted_type,
                            self.converted_bit_size
                        )
                    if 'WRITE' in keyword:
                        self.current_item.write_conversion = GenericConversion(
                            self.proc_text,
                            self.converted_type,
                            self.converted_bit_size
                        )
                    self.building_generic_conversion = False
                else:
                    # Add the current config.line to the conversion being built
                    self.proc_text += parser.line + "\n"
            else:
                # not building generic conversion
                if keyword == 'TABLEFILE':
                    usage = f"{keyword} <File name>"
                    parser.verify_num_parameters(1, 1, usage)
                    table_filename = os.path.join(os.path.dirname(filename), params[0])
                    if not os.path.exists(table_filename):
                        raise parser.error(
                            f"Table file {table_filename} not found",
                            usage,
                        )
                    self._process_file(table_filename)

                elif keyword == 'TABLE':
                    self.finish_packet()
                    self.current_packet = TableParser.parse_table(parser, self.commands, self.warnings)
                    self.definitions[self.current_packet.packet_name] = self.last_config
                    self.current_cmd_or_tlm = self.COMMAND
                    self.default_index = 0

                # Select an existing table for editing
                elif keyword == 'SELECT_TABLE':
                    usage = f"{keyword} <TABLE NAME>"
                    self.finish_packet()
                    parser.verify_num_parameters(1, 1, usage)
                    table_name = params[0].upper()
                    self.current_packet = self.table(table_name)
                    if not self.current_packet:
                        raise parser.error(f"Table {table_name} not found", usage)

                # All the following keywords must have a current packet defined
                elif keyword in ('SELECT_PARAMETER', 'PARAMETER', 'ID_PARAMETER',
                            'ARRAY_PARAMETER', 'APPEND_PARAMETER', 'APPEND_ID_PARAMETER',
                            'APPEND_ARRAY_PARAMETER', 'ALLOW_SHORT', 'HAZARDOUS',
                            'PROCESSOR', 'META', 'DISABLE_MESSAGES', 'DISABLED'):
                    if not self.current_packet:
                        raise parser.error(f"No current packet for {keyword}")
                    self.process_current_packet(parser, keyword, params)

                # All the following keywords must have a current item defined
                elif keyword in ('STATE', 'READ_CONVERSION', 'WRITE_CONVERSION',
                            'POLY_READ_CONVERSION', 'POLY_WRITE_CONVERSION',
                            'SEG_POLY_READ_CONVERSION', 'SEG_POLY_WRITE_CONVERSION',
                            'GENERIC_READ_CONVERSION_START',
                            'GENERIC_WRITE_CONVERSION_START', 'REQUIRED', 'LIMITS',
                            'LIMITS_RESPONSE', 'UNITS', 'FORMAT_STRING', 'DESCRIPTION',
                            'HIDDEN', 'MINIMUM_VALUE', 'MAXIMUM_VALUE', 'DEFAULT_VALUE',
                            'OVERFLOW', 'UNEDITABLE'):
                    if not self.current_item:
                        raise parser.error(f"No current item for {keyword}")
                    self.process_current_item(parser, keyword, params)

                elif keyword == 'DEFAULT':
                    if len(params) != len(self.current_packet.sorted_items):
                        raise parser.error(f"DEFAULT {' '.join(params)} length of {len(params)} doesn't match item length of {len(self.current_packet.sorted_items)}")
                    self.defaults.extend(params)

                else:
                    # blank config.lines will have a nil keyword and should not raise an exception
                    if keyword:  # None is falsy in Python
                        raise parser.error(f"Unknown keyword '{keyword}'")

        # Complete the last defined packet
        self.finish_packet()

    def process_current_packet(self, parser, keyword, params):
        """
        Process the current packet

        Args:
            parser: Configuration parser
            keyword: The keyword being processed
            params: Parameters for the keyword
        """
        try:
            super().process_current_packet(parser, keyword, params)
        except Exception as err:
            if "not found" in str(err):
                raise parser.error(
                    f"{params[0]} not found in table {self.current_packet.table_name}",
                    'SELECT_PARAMETER <PARAMETER NAME>',
                )
            else:
                raise

    def process_current_item(self, parser, keyword, params):
        """
        Overridden method to handle the unique table item parameters: UNEDITABLE
        and HIDDEN.

        Args:
            parser: Configuration parser
            keyword: The keyword being processed
            params: Parameters for the keyword
        """
        super().process_current_item(parser, keyword, params)
        if keyword == 'UNEDITABLE':
            usage = f"{keyword}"
            parser.verify_num_parameters(0, 0, usage)
            self.current_item.editable = False
        elif keyword == 'HIDDEN':
            usage = f"{keyword}"
            parser.verify_num_parameters(0, 0, usage)
            self.current_item.hidden = True

    def start_item(self, parser):
        """
        Start a new table item

        Args:
            parser: Configuration parser
        """
        self.finish_item()
        self.current_item = TableItemParser.parse(parser, self.current_packet, self.warnings)

    def finish_packet(self):
        """
        Finish the packet definition

        If the table is ROW_COLUMN all currently defined items are
        duplicated until the specified number of rows are created.
        """
        if self.current_packet:
            warnings = self.current_packet.check_bit_offsets()
            if len(warnings) > 0:
                raise Exception(f"Overlapping items not allowed in tables.\n{warnings}")

            if self.current_packet.type == "ROW_COLUMN":
                items = self.current_packet.sorted_items.copy()
                self.current_packet.num_columns = len(items)  # Set the number of columns
                for row in range(self.current_packet.num_rows - 1):
                    for item in items:
                        new_item = item.clone()
                        new_item.name = f"{new_item.name}{row + 1}"
                        self.current_packet.append(new_item)

            if self.defaults:
                for index, item in enumerate(self.current_packet.sorted_items):
                    if item.data_type in ("INT", "UINT"):
                        try:
                            # Integer handles hex strings, e.g. 0xDEADBEEF
                            item.default = int(self.defaults[index], 0)
                        except ValueError:
                            value = item.states.get(self.defaults[index])
                            if value is not None:
                                item.default = value
                            else:
                                raise Exception(f"Unknown DEFAULT {self.defaults[index]} for item {item.name}. Valid states are {', '.join(item.states.keys())}.")
                    elif item.data_type == "FLOAT":
                        item.default = float(self.defaults[index])
                    elif item.data_type in ("STRING", "BLOCK"):
                        item.default = self.defaults[index]

            # Reset defaults after processing them
            self.defaults = []

        super().finish_packet()



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError, WaiterError
from openc3.utilities.bucket import Bucket
from openc3.environment import *
import time

aws_arn = OPENC3_AWS_ARN_PREFIX
s3_config = Config(s3={"addressing_style": "path"})
if OPENC3_BUCKET_URL:
    s3_endpoint_url = OPENC3_BUCKET_URL
elif OPENC3_DEVEL:
    s3_endpoint_url = "http://127.0.0.1:9000"
else:
    s3_endpoint_url = "http://openc3-minio:9000"

if OPENC3_CLOUD == "local":
    s3_session = boto3.session.Session(
        aws_access_key_id=OPENC3_BUCKET_USERNAME,
        aws_secret_access_key=OPENC3_BUCKET_PASSWORD,
        region_name="us-east-1",
    )
else:  # AWS
    s3_endpoint_url = f"https://s3.{AWS_REGION}.amazonaws.com"
    s3_session = boto3.session.Session(region_name=AWS_REGION)

class AwsBucket(Bucket):
    CREATE_CHECK_COUNT = 100  # 10 seconds

    def __init__(self):
        # Check whether the session is a real Session or a MockS3
        # print(f"\nAwsBucket INIT session:{s3_session}\n")
        self.client = s3_session.client("s3", endpoint_url=s3_endpoint_url, config=s3_config)

    def create(self, bucket):
        if not self.exist(bucket):
            self.client.create_bucket(Bucket=bucket)
            count = 0
            while True:
                time.sleep(0.1)
                count += 1
                if self.exist(bucket) or count > self.CREATE_CHECK_COUNT:
                    break
        return bucket

    def ensure_public(self, bucket):
        if OPENC3_NO_BUCKET_POLICY is None:
            policy = """{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": [
        "s3:GetBucketLocation",
        "s3:ListBucket"
      ],
      "Effect": "Allow",
      "Principal": {
        "AWS": [
          "*"
        ]
      },
      "Resource": ["""
            policy = policy + f'\n        "{aws_arn}:s3:::{bucket}"'
            policy = (
                policy
                + """
      ],
      "Sid": ""
    },
    {
      "Action": [
        "s3:GetObject"
      ],
      "Effect": "Allow",
      "Principal": {
        "AWS": [
          "*"
        ]
      },
      "Resource": ["""
            )
            policy = policy + f'\n        "{aws_arn}:s3:::{bucket}/*"'
            policy = (
                policy
                + """
      ],
      "Sid": ""
    }
  ]
}"""
            )
            self.client.put_bucket_policy(Bucket=bucket, Policy=policy, ChecksumAlgorithm="SHA256")

    def exist(self, bucket):
        try:
            self.client.head_bucket(Bucket=bucket)
            return True
        except ClientError:
            return False

    def delete(self, bucket):
        if self.exist(bucket):
            self.client.delete_bucket(Bucket=bucket)

    def get_object(self, bucket, key, path=None):
        try:
            if path:
                response = self.client.get_object(Bucket=bucket, Key=key)
                with open(path, "wb") as f:
                    f.write(response["Body"].read())
                return response
            else:
                return self.client.get_object(Bucket=bucket, Key=key)
        # If the key is not found return nil
        except ClientError:
            return None

    def list_objects(self, bucket, prefix=None, max_request=1000, max_total=100_000):
        try:
            result = []
            kw_args = {"Bucket": bucket, "MaxKeys": max_request}
            if prefix:
                kw_args["Prefix"] = prefix
            while True:
                resp = self.client.list_objects_v2(**kw_args)
                if "Contents" in resp:
                    result = result + resp["Contents"]
                if len(result) >= max_total:
                    break
                if not resp["IsTruncated"]:
                    break
                kw_args["ContinuationToken"] = resp["NextContinuationToken"]
            # Array  of objects with key and size methods
            return result
        except ClientError:
            raise Bucket.NotFound(f"Bucket '{bucket}' does not exist.")

    # Lists the files under a specified path
    def list_files(self, bucket, path, only_directories=False, metadata=False):
        try:
            # Trailing slash is important in AWS S3 when listing files
            # See https://docs.aws.amazon.com/sdk-for-ruby/v3/api/Aws/S3/Types/ListObjectsV2Output.html#common_prefixes-instance_method
            if path[-1] != "/":
                path += "/"
            # If we're searching for the root then kill the path or AWS will return nothing
            if path == "/":
                path = None

            result = []
            dirs = []
            files = []
            kw_args = {"Bucket": bucket, "MaxKeys": 1000, "Delimiter": "/"}
            if path:
                kw_args["Prefix"] = path

            while True:
                resp = self.client.list_objects_v2(**kw_args)
                if "CommonPrefixes" in resp:
                    for item in resp["CommonPrefixes"]:
                        # If path was DEFAULT/targets_modified/ then the
                        # results look like DEFAULT/targets_modified/INST/
                        dirs.append(item["Prefix"].split("/")[-2])
                if only_directories:
                    result = dirs
                else:
                    if "Contents" in resp:
                        for aws_item in resp["Contents"]:
                            item = {
                                "name": aws_item["Key"].split("/")[-1],
                                "modified": aws_item["LastModified"],
                                "size": aws_item["Size"],
                            }
                            if metadata:
                                item["metadata"] = self.head_object(bucket=bucket, key=aws_item["Key"])
                            files.append(item)
                    result = (dirs, files)
                if not resp["IsTruncated"]:
                    break
                kw_args["ContinuationToken"] = resp["NextContinuationToken"]
            return result
        except ClientError:
            raise Bucket.NotFound(f"Bucket '{bucket}' does not exist.")

    # get metadata for a specific object
    def head_object(self, bucket, key):
        try:
            return self.client.head_object(Bucket=bucket, Key=key)
        except ClientError:
            raise Bucket.NotFound(f"Object '{bucket}/{key}' does not exist.")

    # put_object fires off the request to store but does not confirm
    def put_object(self, bucket, key, body, content_type=None, cache_control=None, metadata=None):
        kw_args = {
            "Bucket": bucket,
            "Key": key,
            "Body": body,
            "ChecksumAlgorithm": "SHA256",
        }
        if content_type:
            kw_args["ContentType"] = content_type
        if cache_control:
            kw_args["CacheControl"] = cache_control
        if metadata:
            kw_args["Metadata"] = metadata
        return self.client.put_object(**kw_args)

    # @returns [Boolean] Whether the file exists
    def check_object(self, bucket, key, retries=True):
        if retries:
            try:
                s3_object_exists_waiter = self.client.get_waiter("object_exists")
                s3_object_exists_waiter.wait(
                    Bucket=bucket,
                    Key=key,
                    WaiterConfig={"Delay": 0.1, "MaxAttempts": 30},
                )
                return True
            except WaiterError:
                return False
        else:
            try:
                self.head_object(bucket, key)
                return True
            except Bucket.NotFound:
                return False

    def delete_object(self, bucket, key):
        self.client.delete_object(Bucket=bucket, Key=key)

    def delete_objects(self, bucket, keys):
        def method(key):
            return {"Key": key}

        key_list = list(map(method, keys))
        return self.client.delete_objects(Bucket=bucket, Delete={"Objects": key_list})

    def presigned_request(self, bucket, key, method, internal=True):
        prefix = "/" if internal else "/files/"

        fields = None
        if method == "get_object":
            url = self.client.generate_presigned_url("get_object", Params={"Bucket": bucket, "Key": key})
        else:  # put_object
            response = self.client.generate_presigned_post(bucket, key)
            url = response["url"]
            fields = response["fields"]

        url = prefix + "/".join(url.split("/")[3:])
        return {"url": url, "fields": fields}



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import sys
from openc3.microservices.interface_microservice import InterfaceMicroservice
from openc3.system.system import System
from openc3.models.router_status_model import RouterStatusModel
from openc3.topics.router_topic import RouterTopic
from openc3.utilities.thread_manager import ThreadManager

class RouterMicroservice(InterfaceMicroservice):
    def handle_packet(self, packet):
        RouterStatusModel.set(self.interface.as_json(), scope=self.scope)
        if not packet.identified():
            # Need to identify so we can find the target
            identified_packet = System.commands.identify(packet.buffer_no_copy(), self.interface.cmd_target_names)
            if identified_packet:
                packet = identified_packet

        if not packet.defined():
            if packet.target_name and packet.packet_name:
                try:
                    defined_packet = System.commands.packet(packet.target_name, packet.packet_name)
                    defined_packet.received_time = packet.received_time
                    defined_packet.stored = packet.stored
                    defined_packet.buffer = packet.buffer
                    packet = defined_packet
                except RuntimeError:
                    self.logger.warn(f"Error defining packet of {len(packet)} bytes")

        target_name = packet.target_name
        if not target_name:
            target_name = "UNKNOWN"
        target = System.targets.get(target_name)

        try:
            try:
                log_message = True  # Default is True
                # If the packet has the DISABLE_MESSAGES keyword then no messages by default
                if packet.messages_disabled:
                    log_message = False
                # Check if any of the parameters have DISABLE_MESSAGES
                for item in packet.sorted_items:
                    if item.states and item.messages_disabled:
                        value = packet.read_item(item)
                        if item.messages_disabled[value]:
                            log_message = False
                            break

                if log_message:
                    if target and target_name != "UNKNOWN":
                        self.logger.info(System.commands.format(packet, target.ignored_parameters))
                    else:
                        self.logger.warn(
                            f"Unidentified packet of {len(packet.buffer_no_copy())} bytes being routed to target {self.interface.cmd_target_names[0]}"
                        )
            except RuntimeError as error:
                self.logger.error(f"Problem formatting command from router=\n{repr(error)}")

            RouterTopic.route_command(packet, self.interface.cmd_target_names, scope=self.scope)
        except RuntimeError as error:
            self.error = error
            self.logger.error(f"Error routing command from {self.interface.name}\n{repr(error)}")


if os.path.basename(__file__) == os.path.basename(sys.argv[0]):
    RouterMicroservice.class_run()
    ThreadManager.instance().shutdown()
    ThreadManager.instance().join()


# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic
from openc3.utilities.store_queued import EphemeralStoreQueued
from openc3.utilities.time import to_nsec_from_epoch


class TelemetryTopic(Topic):
    @classmethod
    def write_packet(cls, packet, scope, queued=False):
        msg_hash = {
            "time": to_nsec_from_epoch(packet.packet_time),
            "received_time": to_nsec_from_epoch(packet.received_time),
            "stored": str(packet.stored),
            "target_name": packet.target_name,
            "packet_name": packet.packet_name,
            "received_count": packet.received_count,
            "buffer": bytes(packet.buffer_no_copy()),
        }
        if packet.extra:
            msg_hash["extra"] = json.dumps(packet.extra)
        if queued:
            EphemeralStoreQueued.write_topic(
                f"{scope}__TELEMETRY__{{{packet.target_name}}}__{packet.packet_name}",
                msg_hash,
            )
        else:
            Topic.write_topic(
                f"{scope}__TELEMETRY__{{{packet.target_name}}}__{packet.packet_name}",
                msg_hash,
            )



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1963

# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1957

import json
from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.topics.topic import Topic
from openc3.topics.interface_topic import InterfaceTopic
from openc3.topics.decom_interface_topic import DecomInterfaceTopic
from openc3.models.cvt_model import CvtModel
from openc3.models.target_model import TargetModel
from openc3.models.interface_model import InterfaceModel
from openc3.utilities.extract import *

WHITELIST.extend(
    [
        "tlm",
        "tlm_raw",
        "tlm_formatted",
        "tlm_with_units",
        "set_tlm",
        "inject_tlm",
        "override_tlm",
        "get_overrides",
        "normalize_tlm",
        "get_tlm_buffer",
        "get_tlm_packet",
        "get_tlm_values",
        "get_all_tlm",
        "get_all_telemetry",  # DEPRECATED
        "get_all_tlm_names",
        "get_all_telemetry_names",  # DEPRECATED
        "get_all_tlm_item_names",
        "get_tlm",
        "get_telemetry",  # DEPRECATED
        "get_item",
        "subscribe_packets",
        "get_packets",
        "get_tlm_cnt",
        "get_tlm_cnts",
        "get_packet_derived_items",
    ]
)


# Request a telemetry item from a packet.
#
# Accepts two different calling styles:
#   tlm("TGT PKT ITEM")
#   tlm('TGT','PKT','ITEM')
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args [String|Array<String>] See the description for calling style
# @param type [Symbol] Telemetry type, :RAW, :CONVERTED (default), :FORMATTED, or :WITH_UNITS
# @return [Object] The telemetry value formatted as requested
def tlm(*args, type="CONVERTED", cache_timeout=0.1, scope=OPENC3_SCOPE):
    target_name, packet_name, item_name = _tlm_process_args(args, "tlm", cache_timeout=cache_timeout, scope=scope)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    return CvtModel.get_item(target_name, packet_name, item_name, type, cache_timeout, scope)


def tlm_raw(*args, cache_timeout=0.1, scope=OPENC3_SCOPE):
    return tlm(*args, type="RAW", cache_timeout=cache_timeout, scope=scope)


def tlm_formatted(*args, cache_timeout=0.1, scope=OPENC3_SCOPE):
    return tlm(*args, type="FORMATTED", cache_timeout=cache_timeout, scope=scope)


def tlm_with_units(*args, cache_timeout=0.1, scope=OPENC3_SCOPE):
    return tlm(*args, type="WITH_UNITS", cache_timeout=cache_timeout, scope=scope)


# Set a telemetry item in the current value table.
#
# Note: If this is done while OpenC3 is currently receiving telemetry,
# this value could get overwritten at any time. Thus this capability is
# best used for testing or for telemetry items that are not received
# regularly through the target interface.
#
# Accepts two different calling styles:
#   set_tlm("TGT PKT ITEM = 1.0")
#   set_tlm('TGT','PKT','ITEM', 10.0)
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args [String|Array<String>] See the description for calling style
# @param type [Symbol] Telemetry type, :RAW, :CONVERTED (default), :FORMATTED, or :WITH_UNITS
def set_tlm(*args, type="CONVERTED", scope=OPENC3_SCOPE):
    target_name, packet_name, item_name, value = _set_tlm_process_args(args, "set_tlm", scope)
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    CvtModel.set_item(target_name, packet_name, item_name, value, type, scope)


# Injects a packet into the system as if it was received from an interface
#
# @param target_name [String] Target name of the packet
# @param packet_name [String] Packet name of the packet
# @param item_hash [Hash] Hash of item_name and value for each item you want to change from the current value table
# @param type [Symbol] Telemetry type, :RAW, :CONVERTED (default), :FORMATTED, or :WITH_UNITS
def inject_tlm(target_name, packet_name, item_hash=None, type="CONVERTED", scope=OPENC3_SCOPE):
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    target_name = target_name.upper()
    packet_name = packet_name.upper()
    if type not in CvtModel.VALUE_TYPES:
        raise RuntimeError(f"Unknown type '{type}' for {target_name} {packet_name}")

    if item_hash:
        item_hash = {k.upper(): v for k, v in item_hash.items()}
        # Check that the items exist ... exceptions are raised if not
        items = TargetModel.packet_items(target_name, packet_name, item_hash.keys(), scope=scope)
        if type == 'CONVERTED':
            # If the type is converted, check that the item states are valid
            for item_name, item_value in item_hash.items():
                item = next((i for i in items if i['name'] == item_name.upper()), None)
                if item and item.get('states') and item_value not in item['states']:
                    raise RuntimeError(
                        f"Unknown state '{item_value}' for {item['name']}, must be one of {', '.join(item['states'].keys())}"
                    )
    else:
        # Check that the packet exists ... exceptions are raised if not
        TargetModel.packet(target_name, packet_name, scope=scope)

    # See if this target has a tlm interface
    interface_name = None
    for _, interface in InterfaceModel.all(scope).items():
        if target_name in interface["tlm_target_names"]:
            interface_name = interface["name"]
            break

    # Use an interface microservice if it exists, other use the decom microservice
    if interface_name:
        InterfaceTopic.inject_tlm(interface_name, target_name, packet_name, item_hash, type=type, scope=scope)
    else:
        DecomInterfaceTopic.inject_tlm(target_name, packet_name, item_hash, type=type, scope=scope)


# Override the current value table such that a particular item always
# returns the same value (for a given type) even when new telemetry
# packets are received from the target.
#
# Accepts two different calling styles:
#   override_tlm("TGT PKT ITEM = 1.0")
#   override_tlm('TGT','PKT','ITEM', 10.0)
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args The args must either be a string followed by a value or
#   three strings followed by a value (see the calling style in the
#   description).
# @param type [Symbol] Telemetry type, :ALL (default), :RAW, :CONVERTED, :FORMATTED, :WITH_UNITS
def override_tlm(*args, type="ALL", scope=OPENC3_SCOPE):
    target_name, packet_name, item_name, value = _set_tlm_process_args(args, "override_tlm", scope)
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    CvtModel.override(target_name, packet_name, item_name, value, type=type, scope=scope)


# Get the list of CVT overrides
def get_overrides(scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    return CvtModel.overrides(scope=scope)


# Normalize a telemetry item in a packet to its default behavior. Called
# after override_tlm to restore standard processing.
#
# Accepts two different calling styles:
#   normalize_tlm("TGT PKT ITEM")
#   normalize_tlm('TGT','PKT','ITEM')
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args The args must either be a string or three strings
#   (see the calling style in the description).
# @param type [Symbol] Telemetry type, :ALL (default), :RAW, :CONVERTED, :FORMATTED, :WITH_UNITS
#   Also takes :ALL which means to normalize all telemetry types
def normalize_tlm(*args, type="ALL", scope=OPENC3_SCOPE):
    target_name, packet_name, item_name = _tlm_process_args(args, "normalize_tlm", scope=scope)
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    CvtModel.normalize(target_name, packet_name, item_name, type=type, scope=scope)


# Returns the raw buffer for a telemetry packet.
#
# @param target_name [String] Name of the target
# @param packet_name [String] Name of the packet
# @return [Hash] telemetry hash with last telemetry buffer
def get_tlm_buffer(*args, scope=OPENC3_SCOPE):
    target_name, packet_name = _extract_target_packet_names("get_tlm_buffer", *args)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    TargetModel.packet(target_name, packet_name, scope=scope)
    topic = f"{scope}__TELEMETRY__{{{target_name}}}__{packet_name}"
    msg_id, msg_hash = Topic.get_newest_message(topic)
    if msg_id:
        # Decode the keys for user convenience
        return {k.decode(): v for (k, v) in msg_hash.items()}
    return None


def get_tlm_packet(*args, stale_time: int = 30, type: str = "CONVERTED", scope: str = OPENC3_SCOPE):
    """Returns all the values (along with their limits state) for a packet.

    Args:
        target_name (str) Name of the target
        packet_name (str) Name of the packet
        stale_time (int) Time in seconds from Time.now that packet will be marked stale
        type (str) Types returned, :RAW, :CONVERTED (default), :FORMATTED, or :WITH_UNITS

    Return:
        (List[String, Object, Symbol|None]) Returns an Array consisting of
        [item name, item value, item limits state] where the item limits state
        can be one of {OpenC3::Limits::LIMITS_STATES}
    """
    target_name, packet_name = _extract_target_packet_names("get_tlm_packet", *args)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    packet = TargetModel.packet(target_name, packet_name, scope=scope)
    t = _validate_tlm_type(type)
    if t is None:
        raise TypeError(f"Unknown type '{type}' for {target_name} {packet_name}")
    cvt_items = [[target_name, packet_name, item["name"].upper(), type] for item in packet["items"]]
    # This returns an array of arrays containing the value and the limits state:
    # [[0, None], [0, 'RED_LOW'], ... ]
    current_values = CvtModel.get_tlm_values(cvt_items, stale_time=stale_time, scope=scope)
    return [[cvt_items[index][2], item[0], item[1]] for index, item in enumerate(current_values)]


# Returns all the item values (along with their limits state). The items
# can be from any target and packet and thus must be fully qualified with
# their target and packet names.
#
# @param items [Array<String>] Array of items consisting of 'tgt__pkt__item__type'
# @param stale_time [Integer] Time in seconds from Time.now that data will be marked stale
# @return [Array<Object, Symbol>]
#   Array consisting of the item value and limits state
#   given as symbols such as :RED, :YELLOW, :STALE
def get_tlm_values(items, stale_time=30, cache_timeout=0.1, scope=OPENC3_SCOPE):
    if not isinstance(items, list) or len(items) == 0 or not isinstance(items[0], str):
        raise TypeError("items must be array of strings: ['TGT__PKT__ITEM__TYPE', ...]")
    packets = []
    cvt_items = []
    for item in items:
        try:
            target_name, packet_name, item_name, value_type = item.upper().split("__")
        except ValueError:
            raise ValueError("items must be formatted as TGT__PKT__ITEM__TYPE")
        if packet_name == "LATEST":
            packet_name = CvtModel.determine_latest_packet_for_item(target_name, item_name, cache_timeout, scope)
        # Change packet_name in case of LATEST and ensure upcase
        cvt_items.append([target_name, packet_name, item_name, value_type])
        packets.append([target_name, packet_name])
    # Make the array of arrays unique
    packets = [list(x) for x in set(tuple(x) for x in packets)]
    for name in packets:
        authorize(
            permission="tlm",
            target_name=name[0],
            packet_name=name[1],
            scope=scope,
        )
    return CvtModel.get_tlm_values(cvt_items, stale_time, cache_timeout, scope)


# Returns an array of all the telemetry packet hashes
#
# @param target_name [String] Name of the target
# @return [Array<Hash>] Array of all telemetry packet hashes
def get_all_tlm(target_name, scope=OPENC3_SCOPE):
    target_name = target_name.upper()
    authorize(permission="tlm", target_name=target_name, scope=scope)
    return TargetModel.packets(target_name, type="TLM", scope=scope)


# get_all_telemetry is DEPRECATED
get_all_telemetry = get_all_tlm


def get_all_tlm_names(target_name: str, hidden: bool = False, scope: str = OPENC3_SCOPE):
    """Returns an array of all the telemetry packet names

    Args:
        target_name [] Name of the target

    Return:
        List[str] Array of all telemetry packet names
    """
    try:
        packets = get_all_tlm(target_name, scope=scope)
    except RuntimeError:
        packets = []
    names = []
    for packet in packets:
        if hidden:
            names.append(packet["packet_name"])
        else:
            if "hidden" not in packet:
                names.append(packet["packet_name"])
    return names

def get_all_tlm_item_names(target_name: str, hidden: bool = False, scope: str = OPENC3_SCOPE):
    """Returns an array of all the item names for every packet in a target

    Args:
        target_name (str) Name of the target

    Return:
        List[str] Array of all telemetry item names
    """
    authorize(permission="tlm", target_name=target_name, scope=scope)
    try:
        items = TargetModel.all_item_names(target_name, scope=scope)
    except RuntimeError:
        items = []
    return items

# get_all_telemetry_names is DEPRECATED
get_all_telemetry_names = get_all_tlm_names


def get_tlm(*args, scope: str = OPENC3_SCOPE):
    """Returns a telemetry packet hash

    Args:
        scope (str) Name of the scope

    Return:
        (dict) Telemetry packet hash
    """
    target_name, packet_name = _extract_target_packet_names("get_tlm", *args)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    return TargetModel.packet(target_name, packet_name, scope=scope)


# get_telemetry is DEPRECATED
get_telemetry = get_tlm


def get_item(*args, scope: str = OPENC3_SCOPE):
    """Returns a telemetry packet item hash

    Args:
        scope (str) Name of the scope

    Return:
        (dict) Telemetry packet hash
    """
    target_name, packet_name, item_name = _extract_target_packet_item_names("get_item", *args)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    return TargetModel.packet_item(target_name, packet_name, item_name, scope=scope)


# 2x double underscore since __ is reserved
SUBSCRIPTION_DELIMITER = "____"


def subscribe_packets(packets, scope=OPENC3_SCOPE):
    """Subscribe to a list of packets. An ID is returned which is passed to get_packets(id) to return packets.

    Args:
        packets (List[List[str]]) List of consisting of target name, packet name
        scope (str) Name of the scope

    Return:
        (str) ID which should be passed to get_packets
    """
    if not isinstance(packets, list) or not isinstance(packets[0], list):
        raise RuntimeError("packets must be nested array: [['TGT','PKT'],...]")

    result = {}
    for target_name, packet_name in packets:
        target_name = target_name.upper()
        packet_name = packet_name.upper()
        authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
        topic = f"{scope}__DECOM__{{{target_name}}}__{packet_name}"
        id_, _ = Topic.get_newest_message(topic)
        result[topic] = id_ if id_ else "0-0"

    mylist = []
    for k, v in result.items():
        mylist += [k, v]
    return SUBSCRIPTION_DELIMITER.join(mylist)


def get_packets(id, count=1000, scope=OPENC3_SCOPE):
    """Get packets based on ID returned from subscribe_packet.

    Args:
        id (str) ID returned from subscribe_packets or last call to get_packets
        count (int) Maximum number of packets to return from EACH packet stream
        scope ()

    Return:
        [Array<String, Array<Hash>] Array of the ID and array of all packets found
    """
    authorize(permission="tlm", scope=scope)
    # Split the list of topic, ID values and turn it into a hash for easy updates
    items = id.split(SUBSCRIPTION_DELIMITER)
    # Convert it back into a dict to create a lookup
    lookup = dict(zip(items[::2], items[1::2]))
    packets = []
    for topic, topic_id, msg_hash, _ in Topic.read_topics(lookup.keys(), list(lookup.values()), None, count):
        # # Return the original ID and empty array if we didn't get anything
        # for topic, data in xread:
        # for id, msg_hash in data:
        lookup[topic] = topic_id  # save the new ID
        # decode the binary string keys and values to strings
        msg_hash = {k.decode(): v.decode() for (k, v) in msg_hash.items()}
        json_hash = json.loads(msg_hash["json_data"])
        msg_hash.pop("json_data")
        packets.append(msg_hash | json_hash)
    mylist = []
    for k, v in lookup.items():
        mylist += [k, v]
    return (SUBSCRIPTION_DELIMITER.join(mylist), packets)


def get_tlm_cnt(*args, scope: str = OPENC3_SCOPE):
    """Get the receive count for a telemetry packet

    Args:
        args (*args) target_name, packet_name (str) Packet name
        scope (str) Name of the scope

    Return:
        [Numeric] Receive count for the telemetry packet
    """
    target_name, packet_name = _extract_target_packet_names("get_tlm_cnt", *args)
    authorize(permission="system", target_name=target_name, packet_name=packet_name, scope=scope)
    TargetModel.packet(target_name, packet_name, scope=scope)
    return TargetModel.get_telemetry_count(target_name, packet_name, scope=scope)


def get_tlm_cnts(target_packets, scope=OPENC3_SCOPE):
    """Get the transmit counts for telemetry packets

    Args:
        target_packets [Array<Array<String, String>>] Array of arrays containing target_name, packet_name

    Return:
        [Numeric] Transmit count for the command
    """
    authorize(permission="system", scope=scope)
    return TargetModel.get_telemetry_counts(target_packets, scope=scope)

def get_packet_derived_items(*args, scope=OPENC3_SCOPE):
    """Get the list of derived telemetry items for a packet

    Args:
        args (*args) target_name, packet_name (str) Packet name
        scope (str)

    Return:
        # @return [Array<String>] All the ignored telemetry items for a packet.
    """
    target_name, packet_name = _extract_target_packet_names("get_packet_derived_items", *args)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    packet = TargetModel.packet(target_name, packet_name, scope=scope)
    return [item["name"] for item in packet["items"] if item["data_type"] == "DERIVED"]


def _extract_target_packet_names(method_name, *args):
    target_name = None
    packet_name = None
    match len(args):
        case 1:
            try:
                target_name, packet_name = args[0].upper().split()
            except ValueError:
                # We get ValueError if passing not enough parameters
                # The check below for None handles this case
                pass
        case 2:
            target_name = args[0].upper()
            packet_name = args[1].upper()
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    if target_name is None or packet_name is None:
        raise RuntimeError(
            f'ERROR: Both target name and packet name required. Usage: {method_name}("TGT PKT") or {method_name}("TGT", "PKT")'
        )
    return target_name, packet_name


def _extract_target_packet_item_names(method_name, *args):
    target_name = None
    packet_name = None
    item_name = None
    match len(args):
        case 1:
            try:
                target_name, packet_name, item_name = args[0].upper().split()
            except ValueError:  # Thrown when not enough items given
                # Do nothing because below error will catch this
                pass
        case 3:
            target_name = args[0].upper()
            packet_name = args[1].upper()
            item_name = args[2].upper()
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    if target_name is None or packet_name is None or item_name is None:
        raise RuntimeError(
            f'ERROR: Target name, packet name and item name required. Usage: {method_name}("TGT PKT ITEM") or {method_name}("TGT", "PKT", "ITEM")'
        )
    return target_name, packet_name, item_name


def _validate_tlm_type(tlm_type):
    match tlm_type:
        case "RAW":
            return ""
        case "CONVERTED":
            return "C"
        case "FORMATTED":
            return "F"
        case "WITH_UNITS":
            return "U"
    return None


def _tlm_process_args(args, method_name, cache_timeout=0.1, scope=OPENC3_SCOPE):
    match (len(args)):
        case 1:
            target_name, packet_name, item_name = extract_fields_from_tlm_text(args[0])
        case 3:
            target_name = args[0]
            packet_name = args[1]
            item_name = args[2]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    target_name = target_name.upper()
    packet_name = packet_name.upper()
    item_name = item_name.upper()
    if packet_name == "LATEST":
        packet_name = CvtModel.determine_latest_packet_for_item(target_name, item_name, cache_timeout, scope)
    else:
        # Determine if this item exists, it will raise appropriate errors if not
        TargetModel.packet_item(target_name, packet_name, item_name, scope=scope)
    return target_name, packet_name, item_name


def _set_tlm_process_args(args, method_name, scope=OPENC3_SCOPE):
    match len(args):
        case 1:
            target_name, packet_name, item_name, value = extract_fields_from_set_tlm_text(args[0])
        case 4:
            target_name = args[0]
            packet_name = args[1]
            item_name = args[2]
            value = args[3]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    target_name = target_name.upper()
    packet_name = packet_name.upper()
    item_name = item_name.upper()
    # Determine if this item exists, it will raise appropriate errors if not
    TargetModel.packet_item(target_name, packet_name, item_name, scope=scope)

    return target_name, packet_name, item_name, value



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from cbor2 import dumps, loads
from .json_accessor import JsonAccessor


class CborAccessor(JsonAccessor):
    @classmethod
    def class_read_item(cls, item, buffer):
        if item.data_type == "DERIVED":
            return None
        if isinstance(buffer, bytearray):
            parsed = loads(buffer)
        else:
            parsed = buffer
        return super().class_read_item(item, parsed)

    @classmethod
    def class_write_item(cls, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        if isinstance(buffer, bytearray):
            decoded = loads(buffer)
        else:
            decoded = buffer

        super().class_write_item(item, value, decoded)

        if isinstance(buffer, bytearray):
            # buffer[0:] syntax so we copy into the buffer
            buffer[0:] = dumps(decoded)
        return value

    @classmethod
    def class_read_items(cls, items, buffer):
        return super().class_read_items(items, loads(buffer))

    @classmethod
    def class_write_items(cls, items, values, buffer):
        decoded = loads(buffer)
        super().class_write_items(items, values, decoded)
        # buffer[0:] syntax so we copy into the buffer
        buffer[0:] = dumps(decoded)
        return values



# Copyright 2023 OpenC3, Inc
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.:

import json
from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.stash_model import StashModel

WHITELIST.extend(["stash_set", "stash_get", "stash_all", "stash_keys", "stash_delete"])


def stash_set(key, value, scope=OPENC3_SCOPE):
    authorize(permission="script_run", scope=scope)
    return StashModel.set({"name": key, "value": json.dumps(value)}, scope=scope)


def stash_get(key, scope=OPENC3_SCOPE):
    authorize(permission="script_view", scope=scope)
    result = StashModel.get(name=key, scope=scope)
    if result:
        return json.loads(result["value"])
    else:
        return None


def stash_all(scope=OPENC3_SCOPE):
    authorize(permission="script_view", scope=scope)
    all = StashModel.all(scope=scope)
    for key, values in all.items():
        all[key] = json.loads(values["value"])
    return all


def stash_keys(scope=OPENC3_SCOPE):
    authorize(permission="script_view", scope=scope)
    return StashModel.names(scope=scope)


def stash_delete(key, scope=OPENC3_SCOPE):
    authorize(permission="script_run", scope=scope)
    model = StashModel.get_model(name=key, scope=scope)
    if model:
        model.destroy()
    return model



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
#
# A portion of this file was funded by Blue Origin Enterprises, L.P.
# See https://github.com/OpenC3/cosmos/pull/1953

import copy
import base64
import hashlib
import datetime
from .structure import Structure
from .packet_item import PacketItem
from .packet_item_limits import PacketItemLimits
from openc3.conversions.packet_time_formatted_conversion import (
    PacketTimeFormattedConversion,
)
from openc3.conversions.packet_time_seconds_conversion import (
    PacketTimeSecondsConversion,
)
from openc3.conversions.received_count_conversion import ReceivedCountConversion
from openc3.conversions.received_time_formatted_conversion import (
    ReceivedTimeFormattedConversion,
)
from openc3.conversions.received_time_seconds_conversion import (
    ReceivedTimeSecondsConversion,
)
from openc3.utilities.logger import Logger
from openc3.utilities.string import (
    simple_formatted,
    quote_if_necessary,
    class_name_to_filename,
)
from openc3.top_level import get_class_from_module


class Packet(Structure):
    RESERVED_ITEM_NAMES = [
        "PACKET_TIMESECONDS",
        "PACKET_TIMEFORMATTED",
        "RECEIVED_TIMESECONDS",
        "RECEIVED_TIMEFORMATTED",
        "RECEIVED_COUNT",
    ]
    ANY_STATE = "ANY"
    # Valid format types
    VALUE_TYPES = ["RAW", "CONVERTED", "FORMATTED", "WITH_UNITS"]

    def __init__(
        self,
        target_name=None,
        packet_name=None,
        default_endianness="BIG_ENDIAN",
        description=None,
        buffer=None,
        item_class=PacketItem,
    ):
        super().__init__(default_endianness, buffer, item_class)
        # Explicitly call the defined setter methods
        self.target_name = target_name
        self.packet_name = packet_name
        self.description = description
        self.packet_time = None
        self.received_time = None
        self.received_count = 0
        self.id_items = None
        self.hazardous = False
        self.hazardous_description = None
        self.given_values = None
        self.limits_items = []
        self.limits_items_hash = {}
        self.processors = {}
        self.limits_change_callback = None
        self.read_conversion_cache = {}
        # self.short_buffer_allowed = None
        self.raw = None
        self.messages_disabled = False
        self.meta = {}
        self.hidden = False
        self.disabled = False
        self.stored = False
        self.extra = None
        self.cmd_or_tlm = None
        self.template = None
        self.response = None
        self.error_response = None
        self.screen = None
        self.related_items = None
        self.ignore_overlap = False
        self.virtual = False
        self.restricted = False
        self.validator = None

    @property
    def target_name(self):
        return self.__target_name

    @target_name.setter
    def target_name(self, target_name):
        """Sets the target name this packet is associated with. Unidentified packets
        will have target name set to None."""
        if target_name is not None:
            if not isinstance(target_name, str):
                raise TypeError(f"target_name must be a str but is a {target_name.__class__.__name__}")

            self.__target_name = target_name.upper()
        else:
            self.__target_name = None

    @property
    def packet_name(self):
        return self.__packet_name

    @packet_name.setter
    def packet_name(self, packet_name):
        """Sets the packet name. Unidentified packets will have packet name set to None"""
        if packet_name is not None:
            if not isinstance(packet_name, str):
                raise TypeError(f"packet_name must be a str but is a {packet_name.__class__.__name__}")

            self.__packet_name = packet_name.upper()
        else:
            self.__packet_name = None

    @property
    def description(self):
        return self.__description

    @description.setter
    def description(self, description):
        """Sets the packet description"""
        if description is not None:
            if not isinstance(description, str):
                raise TypeError(f"description must be a str but is a {description.__class__.__name__}")

            self.__description = description
        else:
            self.__description = None

    # Returns self.received_time unless a packet item called PACKET_TIME exists that returns
    # a Time object that represents a different timestamp for the packet
    @property
    def packet_time(self):
        item = self.items.get("PACKET_TIME")
        if item is not None:
            return self.read_item(item, "CONVERTED", self.buffer)
        else:
            if self.__packet_time is not None:
                return self.__packet_time
            else:
                return self.received_time

    @packet_time.setter
    def packet_time(self, packet_time):
        self.__packet_time = packet_time

    @property
    def received_time(self):
        return self.__received_time

    @received_time.setter
    def received_time(self, received_time):
        """Sets the received time of the packet"""
        if received_time is not None:
            if not isinstance(received_time, datetime.datetime):
                raise TypeError(f"received_time must be a datetime but is a {received_time.__class__.__name__}")
            self.__received_time = received_time
            self.read_conversion_cache = {}
        else:
            self.__received_time = None

    @property
    def received_count(self):
        return self.__received_count

    @received_count.setter
    def received_count(self, received_count):
        """Sets the packet name. Unidentified packets will have packet name set to None"""
        if not isinstance(received_count, int):
            raise TypeError(f"received_count must be an int but is a {received_count.__class__.__name__}")

        self.__received_count = received_count
        self.read_conversion_cache = {}

    @property
    def virtual(self):
        return self.__virtual

    @virtual.setter
    def virtual(self, virtual):
        self.__virtual = virtual
        if virtual:
            self.hidden = True
            self.disabled = True

    # Tries to identify if a buffer represents the currently defined packet. It:
    # does this by iterating over all the packet items that were created with
    # an ID value and checking whether that ID value is present at the correct
    # location in the buffer.
    #
    # Incorrectly sized buffers will still positively identify if there is:
    # enough data to match the ID values. This is to allow incorrectly sized
    # packets to still be processed as well as possible given the incorrectly
    # sized data.
    #
    # self.param buffer [String] Raw buffer of binary data
    # self.return [Boolean] Whether or not the buffer of data is this packet
    def identify(self, buffer):
        if not buffer:
            return False
        if self.virtual:
            return False
        if not self.id_items:
            return True

        for item in self.id_items:
            try:
                value = self.read_item(item, "RAW", buffer)
            # If the item is not found in the buffer, return None
            except Exception:
                value = None
            if item.id_value != value:
                return False
        return True

    # Reads the values from a buffer at the position of each id_item defined
    # in the packet.
    #
    # self.param buffer [String] Raw buffer of binary data
    # self.return [Array] Array of read id values in order
    def read_id_values(self, buffer):
        if not buffer:
            return []
        if not self.id_items:
            return []

        values = []
        for item in self.id_items:
            try:
                values.append(self.read_item(item, "RAW", buffer))
            # If the item is not found in the buffer, append None
            except Exception:
                values.append(None)
        return values

    # Calculates a unique hashing sum that changes if the parts of the packet configuration change that could affect:
    # the "shape" of the packet.  This value is cached and that packet should not be changed if this method is being used:
    def config_name(self):
        if self.config_name:
            return self.config_name

        string = f"{self.target_name} {self.packet_name}"
        for item in self.sorted_items:
            string += f" ITEM {item.name} {item.bit_offset} {item.bit_size} {item.data_type} {item.array_size} {item.endianness} {item.overflow} {item.states} {item.read_conversion.__class__.__name__ if item.read_conversion else 'NO_CONVERSION'}"

        # Use the hashing algorithm established by System
        digest = hashlib.sha256()
        digest.update(string)
        self.config_name = digest.hexdigest()
        return self.config_name

    @property
    def buffer(self, copy=True):
        return self.allocate_buffer_if_needed()[:]

    @buffer.setter
    def buffer(self, buffer):
        with self.synchronize():
            try:
                self.internal_buffer_equals(buffer)
            # Catch and re-raise the TypeError thrown by internal_buffer_equals
            except TypeError as error:
                raise error
            except ValueError:
                Logger.error(
                    f"{self.target_name} {self.packet_name} buffer ({type(buffer)}) received with actual packet length of {len(buffer)} but defined length of {self.defined_length}"
                )
            self.read_conversion_cache = {}

    # Sets the received time of the packet (without cloning)
    #
    # self.param received_time [Time] Time this packet was received
    def set_received_time_fast(self, received_time):
        self.__received_time = received_time
        with self.synchronize():
            self.read_conversion_cache = {}

    @property
    def hazardous_description(self):
        return self.__hazardous_description

    @hazardous_description.setter
    def hazardous_description(self, hazardous_description):
        """Sets the packet hazardous_description"""
        if hazardous_description is not None:
            if not isinstance(hazardous_description, str):
                raise TypeError(
                    f"hazardous_description must be a str but is a {hazardous_description.__class__.__name__}"
                )

            self.__hazardous_description = hazardous_description
        else:
            self.__hazardous_description = None

    @property
    def given_values(self):
        return self.__given_values

    @given_values.setter
    def given_values(self, given_values):
        """Sets the packet given_values"""
        if given_values is not None:
            if not isinstance(given_values, dict):
                raise TypeError(f"given_values must be a dict but is a {given_values.__class__.__name__}")

            self.__given_values = given_values
        else:
            self.__given_values = None

    @property
    def limits_change_callback(self):
        return self.__limits_change_callback

    @limits_change_callback.setter
    def limits_change_callback(self, limits_change_callback_method):
        self.__limits_change_callback = limits_change_callback_method

    @property
    def template(self):
        return self.__template

    @template.setter
    def template(self, template):
        """Sets the packet template"""
        if template is not None:
            if not isinstance(template, (bytes, bytearray)):
                raise TypeError(f"template must be bytes but is a {template.__class__.__name__}")

            self.__template = template
        else:
            self.__template = None

    # Review bit offset to look for overlapping definitions. This will allow
    # gaps in the packet, but not allow the same bits to be used for multiple
    # variables.
    #
    # self.return [Array<String>] Warning messages for bit definition overlaps
    def check_bit_offsets(self):
        if self.ignore_overlap:
            Logger.debug(f"{self.target_name} {self.packet_name} has IGNORE_OVERLAP so bit overlaps ignored")
            return []

        expected_next_offset = None
        previous_item = None
        warnings = []
        for item in self.sorted_items:
            if expected_next_offset and (item.bit_offset < expected_next_offset) and not item.overlap:
                msg = f"Bit definition overlap at bit offset {item.bit_offset} for packet {self.target_name} {self.packet_name} items {item.name} and {previous_item.name}"
                Logger.warn(msg)
                warnings.append(msg)
            expected_next_offset = Packet.next_bit_offset(item)
            previous_item = item
        return warnings

    # Checks if the packet has any gaps or overlapped items:
    #
    # self.return [Boolean] True if the packet has no gaps or overlapped items:
    def packed(self):
        expected_next_offset = None
        for item in self.sorted_items:
            if expected_next_offset and item.bit_offset != expected_next_offset:
                return False
            expected_next_offset = Packet.next_bit_offset(item)
        return True

    # Returns the bit offset of the next item after the current item if items are packed:
    #
    # self.param item [PacketItem] The item to calculate the next offset for
    # self.return [Integer] Bit Offset of Next Item if Packed:
    @classmethod
    def next_bit_offset(cls, item: PacketItem):
        if item.array_size is not None:
            if item.array_size > 0:
                next_offset = item.bit_offset + item.array_size
            else:
                next_offset = item.array_size
        else:
            next_offset = None
            if item.bit_offset > 0:
                if item.little_endian_bit_field:
                    # Bit offset always refers to the most significant bit of a bitfield
                    bits_remaining_in_last_byte = 8 - (item.bit_offset % 8)
                    if item.bit_size > bits_remaining_in_last_byte:
                        next_offset = item.bit_offset + bits_remaining_in_last_byte
            if not next_offset:
                if item.bit_size > 0:
                    next_offset = item.bit_offset + item.bit_size
                else:
                    next_offset = item.bit_size
        return next_offset

    # Indicates if the packet has been identified:
    # self.return [TrueClass or FalseClass]
    def identified(self):
        return self.target_name is not None and self.packet_name is not None

    # Define an item in the packet. This creates a new instance of the
    # item_class as given in the constructor and adds it to the items hash. It
    # also resizes the buffer to accommodate the new item.
    #
    # self.param name [String] Name of the item. Used by the items hash to retrieve
    #   the item.
    # self.param bit_offset [Integer] Bit offset of the item in the raw buffer
    # self.param bit_size [Integer] Bit size of the item in the raw buffer
    # self.param data_type [Symbol] Type of data contained by the item. This is
    #   dependent on the item_class but by default see StructureItem.
    # self.param array_size [Integer] Set to a non None value if the item is to:
    #   represented as an array.
    # self.param endianness [Symbol] Endianness of this item. By default the
    #   endianness as set in the constructor is used.
    # self.param overflow [Symbol] How to handle value overflows. This is
    #   dependent on the item_class but by default see StructureItem.
    # self.param format_string [String] String to pass to Kernel#sprintf
    # self.param read_conversion [Conversion] Conversion to apply case reading the
    #   item from the packet buffer
    # self.param write_conversion [Conversion] Conversion to apply before writing
    #   the item to the packet buffer
    # self.param id_value [Object] Set to something other than None to indicate that
    #   this item should be used to identify a buffer as this packet. The
    #   id_value should make sense according to the data_type.
    # self.return [PacketItem] The new packet item
    def define_item(
        self,
        name,
        bit_offset,
        bit_size,
        data_type,
        array_size=None,
        endianness=None,
        overflow="ERROR",
        format_string=None,
        read_conversion=None,
        write_conversion=None,
        id_value=None,
    ):
        if endianness is None:
            endianness = self.default_endianness
        item = super().define_item(name, bit_offset, bit_size, data_type, array_size, endianness, overflow)
        return self.packet_define_item(item, format_string, read_conversion, write_conversion, id_value)

    # Add an item to the packet by adding it to the items hash. It also
    # resizes the buffer to accommodate the new item.
    #
    # self.param item [PacketItem] Item to add to the packet
    # self.return [PacketItem] The same packet item
    def define(self, item):
        item = super().define(item)
        self.update_id_items(item)
        self.update_limits_items_cache(item)
        return item

    # Define an item at the end of the packet. This creates a new instance of the
    # item_class as given in the constructor and adds it to the items hash. It
    # also resizes the buffer to accommodate the new item.
    #
    # self.param name (see #define_item)
    # self.param bit_size (see #define_item)
    # self.param data_type (see #define_item)
    # self.param array_size (see #define_item)
    # self.param endianness (see #define_item)
    # self.param overflow (see #define_item)
    # self.param format_string (see #define_item)
    # self.param read_conversion (see #define_item)
    # self.param write_conversion (see #define_item)
    # self.param id_value (see #define_item)
    # self.return (see #define_item)
    def append_item(
        self,
        name,
        bit_size,
        data_type,
        array_size=None,
        endianness=None,
        overflow="ERROR",
        format_string=None,
        read_conversion=None,
        write_conversion=None,
        id_value=None,
    ):
        if endianness is None:
            endianness = self.default_endianness
        item = super().append_item(name, bit_size, data_type, array_size, endianness, overflow)
        return self.packet_define_item(item, format_string, read_conversion, write_conversion, id_value)

    # (see Structure#get_item)
    def get_item(self, name):
        try:
            return super().get_item(name)
        except ValueError as error:
            raise RuntimeError(
                f"Packet item '{self.target_name} {self.packet_name} {name.upper()}' does not exist"
            ) from error

    # Read an item in the packet
    #
    # self.param item [PacketItem] Instance of PacketItem or one of its subclasses
    # self.param value_type [Symbol] How to convert the item before returning it.
    #   Must be one of {VALUE_TYPES}
    # self.param buffer (see Structure#read_item)
    # self.param given_raw Given raw value to optimize
    # self.return The value. 'FORMATTED' and 'WITH_UNITS' values are always returned
    #   as Strings. 'RAW' values will match their data_type. 'CONVERTED' values
    #   can be any type.
    def read_item(self, item, value_type="CONVERTED", buffer=None, given_raw=None):
        if not buffer:
            buffer = self._buffer
        if given_raw:
            # Must clone this since value is returned
            value = copy.copy(given_raw)
        else:
            value = super().read_item(item, "RAW", buffer)
        derived_raw = False
        if item.data_type == "DERIVED" and value_type == "RAW":
            value_type = "CONVERTED"
            derived_raw = True
        match value_type:
            case "RAW":
                # Done above
                pass
            case "CONVERTED" | "FORMATTED" | "WITH_UNITS":
                if item.read_conversion:
                    using_cached_value = False
                    check_cache = buffer == self.buffer
                    if check_cache:
                        with self.synchronize_allow_reads():
                            if self.read_conversion_cache.get(item.name):
                                value = self.read_conversion_cache[item.name]
                                # Make sure cached value is not modified by anyone by creating a deep copy
                                if isinstance(value, str):
                                    value = copy.copy(value)
                                elif isinstance(value, list):
                                    value = value.copy()
                                using_cached_value = True

                    if not using_cached_value:
                        if item.array_size is not None:
                            for index, val in enumerate(value):
                                value[index] = item.read_conversion.call(val, self, buffer)
                        else:
                            value = item.read_conversion.call(value, self, buffer)

                        if check_cache:
                            with self.synchronize_allow_reads():
                                self.read_conversion_cache[item.name] = value

                                # Make sure cached value is not modified by anyone by creating a deep copy
                                if isinstance(value, str):
                                    value = copy.copy(value)
                                elif isinstance(value, list):
                                    value = value.copy()

                # Derived raw values perform read_conversions but nothing else:
                if derived_raw:
                    return value

                # Convert from value to state if possible:
                if item.states:
                    if isinstance(value, list):
                        for index, val in enumerate(value):
                            key = item.states_by_value().get(value[index])
                            if key is not None:
                                value[index] = key
                            elif Packet.ANY_STATE in item.states_by_value().keys():
                                value[index] = item.states_by_value()[Packet.ANY_STATE]
                            else:
                                value[index] = self.apply_format_string_and_units(item, val, value_type)
                    else:
                        key = item.states_by_value().get(value)
                        if key is not None:
                            value = key
                        elif Packet.ANY_STATE in item.states_by_value().keys():
                            value = item.states_by_value()[Packet.ANY_STATE]
                        else:
                            value = self.apply_format_string_and_units(item, value, value_type)
                else:
                    if isinstance(value, list):
                        for index, val in enumerate(value):
                            value[index] = self.apply_format_string_and_units(item, val, value_type)
                    else:
                        value = self.apply_format_string_and_units(item, value, value_type)
            case _:
                # Trim a potentially long string (like if they accidentally pass buffer as value_type):
                if len(str(value_type)) > 10:
                    value_type = str(value_type)[0:10]
                    # Ensure we're not trying to output binary
                    if not value_type.isascii():
                        value_type = simple_formatted(value_type)
                    value_type += "..."
                raise ValueError(
                    f"Unknown value type '{value_type}', must be 'RAW', 'CONVERTED', 'FORMATTED', or 'WITH_UNITS'"
                )
        return value

    # Read a list of items in the structure
    #
    # self.param items [StructureItem] Array of PacketItem or one of its subclasses
    # self.param value_type [Symbol] Value type to read for every item
    # self.param buffer [String] The binary buffer to read the items from
    # self.return Hash of read names and values
    def read_items(self, items, value_type="RAW", buffer=None, raw_value=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        if value_type == "RAW":
            result = super().read_items(items, value_type, buffer)
            # Must handle DERIVED special
            for item in items:
                if item.data_type == "DERIVED":
                    result[item.name] = self.read_item(item, value_type, buffer)
        else:
            result = {}
            for item in items:
                result[item.name] = self.read_item(item, value_type, buffer)
        return result

    # Write an item in the packet
    #
    # self.param item [PacketItem] Instance of PacketItem or one of its subclasses
    # self.param value (see Structure#write_item)
    # self.param value_type (see #read_item)
    # self.param buffer (see Structure#write_item)
    def write_item(self, item, value, value_type="CONVERTED", buffer=None):
        if not buffer:
            buffer = self._buffer
        match value_type:
            case "RAW":
                super().write_item(item, value, value_type, buffer)
            case "CONVERTED":
                if item.states:
                    # Convert from state to value if possible:
                    state_value = item.states.get(str(value).upper())
                    if state_value is not None:
                        value = state_value
                if item.write_conversion:
                    value = item.write_conversion.call(value, self, buffer)
                else:
                    if item.data_type == "DERIVED" and self.accessor.enforce_derived_write_conversion(item):
                        raise RuntimeError(f"Cannot write DERIVED item {item.name} without a write conversion")
                try:
                    super().write_item(item, value, "RAW", buffer)
                except ValueError as error:
                    if item.states and isinstance(value, str) and "invalid literal for" in repr(error):
                        raise ValueError(
                            f"Unknown state '{value}' for {item.name}, must be one of f{', '.join(item.states.keys())}"
                        ) from error
                    else:
                        raise error
            case "FORMATTED" | "WITH_UNITS":
                raise ValueError(f"Invalid value type on write: {value_type}")
            case _:
                # Trim potentially long string (like if they accidentally pass buffer as value_type):
                if len(str(value_type)) > 10:
                    value_type = str(value_type)[0:10]
                    # Ensure we're not trying to output binary
                    if not value_type.isascii():
                        value_type = simple_formatted(value_type)
                    value_type += "..."
                raise ValueError(
                    f"Unknown value type '{value_type}', must be 'RAW', 'CONVERTED', 'FORMATTED', or 'WITH_UNITS'"
                )
        with self.synchronize():
            self.read_conversion_cache = {}

    # Write values to the buffer based on the item definitions
    #
    # self.param items [StructureItem] Array of StructureItem or one of its subclasses
    # self.param value [Object] Array of values based on the item definitions.
    # self.param value_type [Symbol] Value type of each item to write
    # self.param buffer [String] The binary buffer to write the values to
    def write_items(self, items, values, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        if value_type == "RAW":
            return super().write_items(items, values, value_type, buffer)
        else:
            for index, item in enumerate(items):
                self.write_item(item, values[index], value_type, buffer)
        return buffer

    # Read an item in the packet by name
    #
    # self.param name [String] Name of the item to read
    # self.param value_type (see #read_item)
    # self.param buffer (see #read_item)
    # self.return (see #read_item)
    def read(self, name, value_type="CONVERTED", buffer=None):
        if not buffer:
            buffer = self._buffer
        return super().read(name, value_type, buffer)

    # Write an item in the packet by name
    #
    # self.param name [String] Name of the item to write
    # self.param value (see #write_item)
    # self.param value_type (see #write_item)
    # self.param buffer (see #write_item)
    def write(self, name, value, value_type="CONVERTED", buffer=None):
        if not buffer:
            buffer = self._buffer
        super().write(name, value, value_type, buffer)

    # Read all items in the packet into an array of arrays
    #   [[item name, item value], ...]
    #
    # self.param value_type (see #read_item)
    # self.param buffer (see Structure#read_all)
    # self.param top (See Structure#read_all)
    # self.return (see Structure#read_all)
    def read_all(self, value_type="CONVERTED", buffer=None, top=True):
        if not buffer:
            buffer = self._buffer
        return super().read_all(value_type, buffer, top)

    # Read all items in the packet into an array of arrays
    #   [[item name, item value], [item limits state], ...]
    #
    # self.param value_type (see #read_all)
    # self.param buffer (see #read_all)
    # self.return [Array<String, Object, Symbol|None>] Returns an Array consisting
    #   of [item name, item value, item limits state] where the item limits
    #   state can be one of {OpenC3:'L'imits='LIMITS_STATES'}
    def read_all_with_limits_states(self, value_type="CONVERTED", buffer=None):
        if not buffer:
            buffer = self._buffer
        result = None
        with self.synchronize_allow_reads(True):
            result = self.read_all(value_type, buffer, False)
            for item in result:
                item.append(self.items[item[0]].limits.state)
        return result

    # Create a string that shows the name and value of each item in the packet
    #
    # self.param value_type (see #read_item)
    # self.param indent (see Structure#formatted)
    # self.param buffer (see Structure#formatted)
    # self.param ignored (see Structure#ignored)
    # self.return (see Structure#formatted)
    def formatted(self, value_type="CONVERTED", indent=0, buffer=None, ignored=None):
        if not buffer:
            buffer = self._buffer
        return super().formatted(value_type, indent, buffer, ignored)

    # Restore all items in the packet to their default value
    #
    # self.param buffer [String] Raw buffer of binary data
    # self.param skip_item_names [Array] Array of item names to skip
    # self.param use_templase [Boolean] Apply template before setting defaults (or not)
    def restore_defaults(self, buffer=None, skip_item_names=None, use_template=True):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        if skip_item_names:
            upcase_skip_item_names = [name.upper() for name in skip_item_names]
        if self.template and use_template:
            # Set both the internal buffer and our local copy
            self.buffer = self.template
            buffer = self._buffer
        for item in self.sorted_items:
            if item.name in Packet.RESERVED_ITEM_NAMES:
                continue
            if item.default is not None:
                if not (skip_item_names and item.name in upcase_skip_item_names):
                    self.write_item(item, item.default, "CONVERTED", buffer)

    # Define the reserved items on the current telemetry packet
    def define_reserved_items(self):
        item = self.define_item(
            "PACKET_TIMESECONDS",
            0,
            0,
            "DERIVED",
            None,
            self.default_endianness,
            "ERROR",
            "%0.6f",
            PacketTimeSecondsConversion(),
        )
        item.description = "OpenC3 Packet Time (UTC, Floating point, Unix epoch)"
        item = self.define_item(
            "PACKET_TIMEFORMATTED",
            0,
            0,
            "DERIVED",
            None,
            self.default_endianness,
            "ERROR",
            None,
            PacketTimeFormattedConversion(),
        )
        item.description = "OpenC3 Packet Time (Local time zone, Formatted string)"
        item = self.define_item(
            "RECEIVED_TIMESECONDS",
            0,
            0,
            "DERIVED",
            None,
            self.default_endianness,
            "ERROR",
            "%0.6f",
            ReceivedTimeSecondsConversion(),
        )
        item.description = "OpenC3 Received Time (UTC, Floating point, Unix epoch)"
        item = self.define_item(
            "RECEIVED_TIMEFORMATTED",
            0,
            0,
            "DERIVED",
            None,
            self.default_endianness,
            "ERROR",
            None,
            ReceivedTimeFormattedConversion(),
        )
        item.description = "OpenC3 Received Time (Local time zone, Formatted string)"
        item = self.define_item(
            "RECEIVED_COUNT",
            0,
            0,
            "DERIVED",
            None,
            self.default_endianness,
            "ERROR",
            None,
            ReceivedCountConversion(),
        )
        item.description = "OpenC3 packet received count"

    # Reset the packet to just derived items
    def clear_all_non_derived_items(self):
        self.defined_length = 0
        self.defined_length_bits = 0
        self.pos_bit_size = 0
        self.neg_bit_size = 0
        self.fixed_size = True
        self.short_buffer_allowed = False
        self.id_items = None
        self.limits_items = None
        new_items = {}
        new_sorted_items = []
        for name, item in self.items.items():
            if item.data_type == "DERIVED":
                new_items[name] = item
        for item in self.sorted_items:
            if item.data_type == "DERIVED":
                new_sorted_items.append(item)
        self.items = new_items
        self.sorted_items = new_sorted_items
        self.config_name = None

    # Enable limits on an item by name
    #
    # self.param name [String] Name of the item to enable limits
    def enable_limits(self, name):
        self.get_item(name).limits.enabled = True

    # Disable limits on an item by name
    #
    # self.param name [String] Name of the item to disable limits
    def disable_limits(self, name):
        item = self.get_item(name)
        item.limits.enabled = False
        if not item.limits.state == "STALE":
            old_limits_state = item.limits.state
            item.limits.state = None
            if self.limits_change_callback is not None:
                self.limits_change_callback(self, item, old_limits_state, None, False)

    # Add an item to the limits items cache if necessary.:
    # You MUST call this after adding limits to an item
    # This is an optimization so we don't have to iterate through all the items case
    # checking for limits.
    def update_limits_items_cache(self, item):
        if item.limits.values or item.state_colors:
            if not self.limits_items_hash.get(item.name):
                self.limits_items.append(item)
                self.limits_items_hash[item.name] = True

    # Return an array of arrays indicating all items in the packet that are out of limits
    #   [[target name, packet name, item name, item limits state], ...]
    #
    # self.return [Array<Array<String, String, String, Symbol>>]
    def out_of_limits(self):
        items = []
        if len(self.limits_items) == 0:
            return items

        for item in self.limits_items:
            if item.limits.enabled and item.limits.state and item.limits.state in PacketItemLimits.OUT_OF_LIMITS_STATES:
                items.append([self.target_name, self.packet_name, item.name, item.limits.state])
        return items

    # Check all the items in the packet against their defined limits. Update
    # their internal limits state and persistence and call the
    # limits_change_callback as necessary.
    #
    # self.param limits_set [Symbol] Which limits set to check the item values
    #   against.
    # self.param ignore_persistence [Boolean] Whether to ignore persistence case
    #   checking for out of limits
    def check_limits(self, limits_set="DEFAULT", ignore_persistence=False):
        if not self.limits_items or len(self.limits_items) == 0:
            return

        for item in self.limits_items:
            # Verify limits monitoring is enabled for this item
            if item.limits.enabled:
                value = self.read_item(item)

                # Handle state monitoring and value monitoring differently
                if item.states is not None:
                    self.handle_limits_states(item, value)
                elif item.limits.values is not None:
                    self.handle_limits_values(item, value, limits_set, ignore_persistence)

    # Reset temporary packet data
    # This includes packet received time, received count, and processor state
    def reset(self):
        # The SYSTEM META packet is a special case that does not get reset
        if self.target_name == "SYSTEM" and self.packet_name == "META":
            return

        self.received_time = None
        self.received_count = 0
        self.stored = False
        self.extra = None
        with self.synchronize():
            self.read_conversion_cache = {}
        if not self.processors:
            return

        for _, processor in self.processors:
            processor.reset

    # Make a light weight clone of this packet. This only creates a new buffer
    # of data and clones the processors. The defined packet items are the same.
    #
    # self.return [Packet] A copy of the current packet with a new underlying
    #   buffer of data and processors
    def clone(self):
        packet = super().clone()
        if self.processors is not None:
            packet.processors = copy.deepcopy(packet.processors)
        packet.read_conversion_cache = {}
        if packet.extra is not None:
            packet.extra = copy.deepcopy(packet.extra)
        return packet

    def update_id_items(self, item):
        if item.id_value is not None:
            if self.id_items is None:
                self.id_items = []
            # Add to Id Items
            if len(self.id_items) > 0:
                last_item = self.id_items[-1]
                self.id_items.append(item)
                # If the current item or last item have a negative offset then we have
                # to re-sort. We also re-sort if the current item is less than the last:
                # item because we are inserting.
                if last_item.bit_offset <= 0 or item.bit_offset <= 0 or item.bit_offset < last_item.bit_offset:
                    self.id_items.sort()
            else:
                self.id_items.append(item)
        return item

    def to_config(self, cmd_or_tlm):
        config = ""

        if cmd_or_tlm == "TELEMETRY":
            config += f'TELEMETRY {quote_if_necessary(self.target_name)} {quote_if_necessary(self.packet_name)} {self.default_endianness} "{self.description}"\n'
        else:
            config += f'COMMAND {quote_if_necessary(self.target_name)} {quote_if_necessary(self.packet_name)} {self.default_endianness} "{self.description}"\n'
        if self.accessor.__class__.__name__ != "BinaryAccessor":
            config += f"  ACCESSOR {self.accessor.__class__.__name__}\n"
        if self.validator:
            config += f"  VALIDATOR {self.validator.__class__.__name__}\n"
        # TODO: Add TEMPLATE_ENCODED so this can always be done inline regardless of content
        if self.template:
            config += f"  TEMPLATE '{self.template}'\n"
        if self.short_buffer_allowed:
            config += "  ALLOW_SHORT\n"
        if self.hazardous:
            config += f"  HAZARDOUS {quote_if_necessary(self.hazardous_description)}\n"
        if self.messages_disabled:
            config += "  DISABLE_MESSAGES\n"
        if self.virtual:
            config += "  VIRTUAL\n"
        elif self.disabled:
            config += "  DISABLED\n"
        elif self.hidden:
            config += "  HIDDEN\n"
        if self.restricted:
            config += "  RESTRICTED\n"

        if self.processors:
            for _, processor in self.processors:
                config += processor.to_config

        if self.meta:
            for key, values in self.meta:
                vals = " ".join([quote_if_necessary(a) for a in values])
                config += f"  META {quote_if_necessary(key)} {vals}\n"

        # Items with derived items last
        for item in self.sorted_items:
            if item.data_type != "DERIVED":
                config += item.to_config(cmd_or_tlm, self.default_endianness)
        for item in self.sorted_items:
            if item.data_type == "DERIVED":
                if item.name not in Packet.RESERVED_ITEM_NAMES:
                    config += item.to_config(cmd_or_tlm, self.default_endianness)

        if self.response:
            config += f"  RESPONSE {quote_if_necessary(self.response[0])} {quote_if_necessary(self.response[1])}\n"

        if self.error_response:
            config += f"  ERROR_RESPONSE {quote_if_necessary(self.error_response[0])} {quote_if_necessary(self.error_response[1])}\n"

        if self.screen:
            config += f"  SCREEN {quote_if_necessary(self.screen[0])} {quote_if_necessary(self.screen[1])}\n"

        if self.related_items:
            for related_item in self.related_items:
                config += f"  RELATED_ITEM {quote_if_necessary(related_item[0])} {quote_if_necessary(related_item[1])} {quote_if_necessary(related_item[2])}\n"

        if self.ignore_overlap:
            config += "  IGNORE_OVERLAP\n"

        return config

    def as_json(self):
        config = {}
        config["target_name"] = self.target_name
        config["packet_name"] = self.packet_name
        config["endianness"] = self.default_endianness
        config["description"] = self.description
        if self.short_buffer_allowed:
            config["short_buffer_allowed"] = True
        if self.hazardous:
            config["hazardous"] = True
        if self.hazardous_description:
            config["hazardous_description"] = self.hazardous_description
        if self.messages_disabled:
            config["messages_disabled"] = True
        if self.disabled:
            config["disabled"] = True
        if self.hidden:
            config["hidden"] = True
        if self.virtual:
            config["virtual"] = True
        if self.restricted:
            config["restricted"] = True
        config["accessor"] = self.accessor.__class__.__name__
        # config["accessor_args"] = self.accessor.args
        if self.validator:
            config["validator"] = self.validator.__class__.__name__
        if self.template:
            config["template"] = base64.b64encode(self.template)

        if self.processors:
            processors = []
            for _, processor in self.processors():
                processors.append(processor.as_json())
            config["processors"] = processors

        if self.meta:
            config["meta"] = self.meta

        items = []
        config["items"] = items
        # Items with derived items last
        for item in self.sorted_items:
            if item.data_type != "DERIVED":
                items.append(item.as_json())

        for item in self.sorted_items:
            if item.data_type == "DERIVED":
                items.append(item.as_json())

        if self.response:
            config["response"] = self.response
        if self.error_response:
            config["error_response"] = self.error_response
        if self.screen:
            config["screen"] = self.screen
        if self.related_items:
            config["related_items"] = self.related_items
        if self.ignore_overlap:
            config["ignore_overlap"] = self.ignore_overlap

        return config

    @classmethod
    def from_json(cls, hash):
        endianness = hash.get("endianness")
        packet = Packet(hash["target_name"], hash["packet_name"], endianness, hash["description"])
        packet.short_buffer_allowed = hash.get("short_buffer_allowed")
        packet.hazardous = hash.get("hazardous")
        packet.hazardous_description = hash.get("hazardous_description")
        packet.messages_disabled = hash.get("messages_disabled")
        packet.disabled = hash.get("disabled")
        packet.hidden = hash.get("hidden")
        packet.virtual = hash.get("virtual")
        packet.restricted = hash.get("restricted")
        if "accessor" in hash:
            try:
                filename = class_name_to_filename(hash["accessor"])
                accessor = get_class_from_module(f"openc3.accessors.{filename}", hash["accessor"])
                if hash.get("accessor_args") and len(hash["accessor_args"]) > 0:
                    packet.accessor = accessor(packet, *hash["accessor_args"])
                else:
                    packet.accessor = accessor(packet)
            except RuntimeError as error:
                Logger.error(
                    f"{packet.target_name} {packet.packet_name} accessor of {hash['accessor']} could not be found due to {repr(error)}"
                )
        if "validator" in hash:
            try:
                filename = class_name_to_filename(hash["validator"])
                validator = get_class_from_module(filename, hash["validator"])
                packet.validator = validator(packet)
            except RuntimeError as error:
                Logger.error(
                    f"{packet.target_name} {packet.packet_name} validator of {hash['validator']} could not be found due to {repr(error)}"
                )
        if "template" in hash:
            packet.template = base64.b64decode(hash["template"])
        packet.meta = hash.get("meta")
        # Can't convert processors
        for item in hash["items"]:
            packet.define(PacketItem.from_json(item))

        if "response" in hash:
            packet.response = hash["response"]
        if "error_response" in hash:
            packet.error_response = hash["error_response"]
        if "screen" in hash:
            packet.screen = hash["screen"]
        if "related_items" in hash:
            packet.related_items = hash["related_items"]
        if "ignore_overlap" in hash:
            packet.ignore_overlap = hash["ignore_overlap"]

        return packet

    def decom(self):
        # Read all the RAW at once because this could be optimized by the accessor
        json_hash = self.read_items(self.sorted_items)

        # Decom extra into the values (overrides packet items)
        if self.extra is not None:
            for key, value in self.extra.items():
                json_hash[key.upper()] = value

        # Now read all other value types - no accessor required
        for item in self.sorted_items:
            given_raw = json_hash[item.name]
            if item.states or (item.read_conversion and item.data_type != "DERIVED"):
                json_hash[f"{item.name}__C"] = self.read_item(item, "CONVERTED", self.buffer, given_raw)
            if item.format_string:
                json_hash[f"{item.name}__F"] = self.read_item(item, "FORMATTED", self.buffer, given_raw)
            if item.units:
                json_hash[f"{item.name}__U"] = self.read_item(item, "WITH_UNITS", self.buffer, given_raw)
            limits_state = item.limits.state
            if limits_state:
                json_hash[f"{item.name}__L"] = limits_state

        return json_hash

    # Performs packet specific processing on the packet.
    # Intended to only be run once for each packet received
    def process(self, buffer=None):
        if not buffer:
            buffer = self._buffer
        if not self.processors:
            return

        for _, processor in self.processors.items():
            processor.call(self, buffer)

    def handle_limits_states(self, item, value):
        # Retrieve limits state for the given value
        limits_state = item.state_colors.get(value)

        if item.limits.state != limits_state:  # PacketItemLimits state has changed
            # Save old limits state
            old_limits_state = item.limits.state
            # Update to new limits state
            item.limits.state = limits_state

            if self.limits_change_callback is not None:
                if item.limits.state is None:
                    self.limits_change_callback(self, item, old_limits_state, value, False)
                else:
                    self.limits_change_callback(self, item, old_limits_state, value, True)

    def handle_limits_values(self, item, value, limits_set, ignore_persistence):
        limits = None

        # Retrieve limits settings for the specified limits_set
        limits = item.limits.values.get(limits_set)

        # Use the default limits set if limits aren't specified for the:
        # particular limits set
        if limits is None:
            limits = item.limits.values["DEFAULT"]

        # Extract limits from array
        red_low = limits[0]
        yellow_low = limits[1]
        yellow_high = limits[2]
        red_high = limits[3]
        if len(limits) > 4:
            green_low = limits[4]
            green_high = limits[5]
        else:
            green_low = None
            green_high = None
        limits_state = None

        # Determine the limits_state based on the limits values and the current
        # value of the item
        if value > yellow_low:
            if value < yellow_high:
                if green_low:
                    if value < green_high:
                        if value > green_low:
                            limits_state = "BLUE"
                        else:
                            limits_state = "GREEN_LOW"
                    else:
                        limits_state = "GREEN_HIGH"
                else:
                    limits_state = "GREEN"
            elif value < red_high:
                limits_state = "YELLOW_HIGH"
            else:
                limits_state = "RED_HIGH"
        else:  # value <= yellow_low
            if value > red_low:
                limits_state = "YELLOW_LOW"
            else:
                limits_state = "RED_LOW"

        if item.limits.state != limits_state:  # limits state has changed
            # Save old limits state for use in the callback
            old_limits_state = item.limits.state

            item.limits.persistence_count += 1

            # Check for item to achieve its persistence which means we
            # have to update the state and call the callback
            # Note case going back to green (or blue) persistence is ignored
            if (item.limits.persistence_count >= item.limits.persistence_setting) or ignore_persistence:
                item.limits.state = limits_state

                # Additional actions for limits change
                if self.limits_change_callback is not None:
                    self.limits_change_callback(self, item, old_limits_state, value, True)

                # Clear persistence since we've entered a new state
                item.limits.persistence_count = 0
        else:  # limits state has not changed so clear persistence
            item.limits.persistence_count = 0

    def apply_format_string_and_units(self, item, value, value_type):
        if value_type == "FORMATTED" or value_type == "WITH_UNITS":
            if item.format_string and value is not None:
                value = f"{item.format_string}" % value
            else:
                value = str(value)
        if value_type == "WITH_UNITS" and item.units:
            value += " " + item.units
        return value

    def packet_define_item(self, item, format_string, read_conversion, write_conversion, id_value):
        item.format_string = format_string
        item.read_conversion = read_conversion
        item.write_conversion = write_conversion

        # Change id_value to the correct type
        if id_value is not None:
            item.id_value = id_value
            self.update_id_items(item)
        return item



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os

from openc3.__version__ import __title__, __version__

_openc3_api_schema = "OPENC3_API_SCHEMA"
_openc3_api_hostname = "OPENC3_API_HOSTNAME"
_openc3_api_port = "OPENC3_API_PORT"
_openc3_api_timeout = "OPENC3_API_TIMEOUT"
_openc3_script_api_schema = "OPENC3_SCRIPT_API_SCHEMA"
_openc3_script_api_hostname = "OPENC3_SCRIPT_API_HOSTNAME"
_openc3_script_api_port = "OPENC3_SCRIPT_API_PORT"
_openc3_script_api_timeout = "OPENC3_SCRIPT_API_TIMEOUT"
_openc3_scope = "OPENC3_SCOPE"
_openc3_api_password = "OPENC3_API_PASSWORD"
_openc3_log_level = "OPENC3_LOG_LEVEL"
_openc3_no_store = "OPENC3_NO_STORE"
_openc3_user_agent = "OPENC3_USER_AGENT"
_openc3_redis_hostname = "OPENC3_REDIS_HOSTNAME"
_openc3_redis_port = "OPENC3_REDIS_PORT"
_openc3_redis_username = "OPENC3_REDIS_USERNAME"
_openc3_redis_password = "OPENC3_REDIS_PASSWORD"
_openc3_redis_ephemeral_hostname = "OPENC3_REDIS_EPHEMERAL_HOSTNAME"
_openc3_redis_ephemeral_port = "OPENC3_REDIS_EPHEMERAL_PORT"
_openc3_cloud = "OPENC3_CLOUD"
_aws_region = "AWS_REGION"
_openc3_aws_arn_prefix = "OPENC3_AWS_ARN_PREFIX"
_openc3_bucket_url = "OPENC3_BUCKET_URL"
_openc3_bucket_username = "OPENC3_BUCKET_USERNAME"
_openc3_bucket_password = "OPENC3_BUCKET_PASSWORD"
_openc3_devel = "OPENC3_DEVEL"
_openc3_full_backtrace = "OPENC3_FULL_BACKTRACE"
_openc3_config_bucket = "OPENC3_CONFIG_BUCKET"
_openc3_logs_bucket = "OPENC3_LOGS_BUCKET"
_openc3_tools_bucket = "OPENC3_TOOLS_BUCKET"
_openc3_local_mode = "OPENC3_LOCAL_MODE"
_openc3_local_mode_path = "OPENC3_LOCAL_MODE_PATH"
_openc3_no_bucket_policy = "OPENC3_NO_BUCKET_POLICY"
_openc3_log_stderr = "OPENC3_LOG_STDERR"

# The following variables are only used with COSMOS Enterprise
_openc3_api_user = "OPENC3_API_USER"
_openc3_api_client = "OPENC3_API_CLIENT"
_openc3_api_token = "OPENC3_API_TOKEN"
_openc3_keycloak_realm = "OPENC3_KEYCLOAK_REALM"
_openc3_keycloak_url = "OPENC3_KEYCLOAK_URL"
_openc3_redis_cluster = "OPENC3_REDIS_CLUSTER"

OPENC3_API_SCHEMA = os.environ.get(_openc3_api_schema, "http")
OPENC3_API_HOSTNAME = os.environ.get(_openc3_api_hostname, "openc3-cosmos-cmd-tlm-api")
try:
    OPENC3_API_PORT = int(os.environ.get(_openc3_api_port))
except TypeError:
    OPENC3_API_PORT = 2901
try:
    OPENC3_API_TIMEOUT = float(os.environ.get(_openc3_api_timeout))
except TypeError:
    OPENC3_API_TIMEOUT = 1.0

OPENC3_SCRIPT_API_SCHEMA = os.environ.get(_openc3_script_api_schema, "http")
OPENC3_SCRIPT_API_HOSTNAME = os.environ.get(_openc3_script_api_hostname, "openc3-cosmos-script-runner-api")
try:
    OPENC3_SCRIPT_API_PORT = int(os.environ.get(_openc3_script_api_port))
except TypeError:
    OPENC3_SCRIPT_API_PORT = 2902
try:
    OPENC3_SCRIPT_API_TIMEOUT = float(os.environ.get(_openc3_script_api_timeout))
except TypeError:
    OPENC3_SCRIPT_API_TIMEOUT = 5.0

OPENC3_REDIS_HOSTNAME = os.environ.get(_openc3_redis_hostname, "openc3-redis")
try:
    OPENC3_REDIS_PORT = int(os.environ.get(_openc3_redis_port))
except TypeError:
    OPENC3_REDIS_PORT = 6379
OPENC3_REDIS_USERNAME = os.environ.get(_openc3_redis_username)
OPENC3_REDIS_PASSWORD = os.environ.get(_openc3_redis_password)
OPENC3_REDIS_EPHEMERAL_HOSTNAME = os.environ.get(_openc3_redis_ephemeral_hostname, "openc3-redis-ephemeral")
try:
    OPENC3_REDIS_EPHEMERAL_PORT = int(os.environ.get(_openc3_redis_ephemeral_port))
except TypeError:
    OPENC3_REDIS_EPHEMERAL_PORT = 6380
OPENC3_CLOUD = os.environ.get(_openc3_cloud, "local")
AWS_REGION = os.environ.get(_aws_region)
OPENC3_AWS_ARN_PREFIX = os.environ.get(_openc3_aws_arn_prefix, "arn:aws")
OPENC3_BUCKET_URL = os.environ.get(_openc3_bucket_url)
OPENC3_BUCKET_USERNAME = os.environ.get(_openc3_bucket_username)
OPENC3_BUCKET_PASSWORD = os.environ.get(_openc3_bucket_password)
OPENC3_DEVEL = os.environ.get(_openc3_devel)
OPENC3_FULL_BACKTRACE = os.environ.get(_openc3_full_backtrace)
OPENC3_CONFIG_BUCKET = os.environ.get(_openc3_config_bucket)
OPENC3_LOGS_BUCKET = os.environ.get(_openc3_logs_bucket)
OPENC3_TOOLS_BUCKET = os.environ.get(_openc3_tools_bucket)
OPENC3_LOCAL_MODE = os.environ.get(_openc3_local_mode)
OPENC3_LOCAL_MODE_PATH = os.environ.get(_openc3_local_mode_path)
OPENC3_NO_BUCKET_POLICY = os.environ.get(_openc3_no_bucket_policy)
OPENC3_LOG_STDERR = os.environ.get(_openc3_log_stderr)

OPENC3_SCOPE = os.environ.get(_openc3_scope, "DEFAULT")
OPENC3_API_PASSWORD = os.environ.get(_openc3_api_password)
OPENC3_LOG_LEVEL = os.environ.get(_openc3_log_level, "INFO")
OPENC3_NO_STORE = os.environ.get(_openc3_no_store)
OPENC3_API_USER = os.environ.get(_openc3_api_user)
OPENC3_API_CLIENT = os.environ.get(_openc3_api_client, "api")
OPENC3_API_TOKEN = os.environ.get(_openc3_api_token)
OPENC3_KEYCLOAK_REALM = os.environ.get(_openc3_keycloak_realm, "openc3")
OPENC3_KEYCLOAK_URL = os.environ.get(_openc3_keycloak_url, "http://openc3-keycloak/auth")
OPENC3_REDIS_CLUSTER = os.environ.get(_openc3_redis_cluster)

# User agent used by all
_openc3_default_user_agent = [
    f"{__title__}:{__version__}",
]

if OPENC3_API_USER is not None:
    _openc3_default_user_agent[0] += f":({OPENC3_API_USER})"

if os.name == "nt":
    _openc3_default_user_agent.append(f"{os.environ.get('COMPUTERNAME')}:{os.environ.get('USERNAME')}")
else:
    _openc3_default_user_agent.append(f"{os.environ.get('HOSTNAME')}:{os.environ.get('USER')}")

OPENC3_USER_AGENT = os.environ.get(_openc3_user_agent, " ".join(_openc3_default_user_agent))



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

# You can quickly setup an unauthenticated MQTT server in Docker with
# docker run -it -p 1883:1883 eclipse-mosquitto:2.0.15 mosquitto -c /mosquitto-no-auth.conf
# You can also test against encrypted and authenticated servers at https://test.mosquitto.org/

import tempfile
from openc3.interfaces.stream_interface import StreamInterface
from openc3.config.config_parser import ConfigParser
from openc3.streams.mqtt_stream import MqttStream


# Base class for interfaces that send and receive messages over MQTT
class MqttStreamInterface(StreamInterface):
    # @param hostname [String] MQTT server to connect to
    # @param port [Integer] MQTT port
    def __init__(self, hostname, port=1883, ssl=False, write_topic=None, read_topic=None, protocol_type=None, protocol_args=[]):
        super().__init__(protocol_type, protocol_args)
        self.hostname = hostname
        self.port = int(port)
        self.ssl = ConfigParser.handle_true_false(ssl)
        self.write_topic = ConfigParser.handle_none(write_topic)
        self.read_topic = ConfigParser.handle_none(read_topic)
        self.ack_timeout = 5.0
        self.username = None
        self.password = None
        self.cert = None
        self.key = None
        self.ca_file = None
        self.keyfile_password = None

    def connection_string(self):
        result = f"{self.hostname}:{self.port} (ssl: {self.ssl})"
        if self.write_topic is not None:
            result += f" write topic: {self.write_topic}"
        if self.read_topic is not None:
            result += f" read topic: {self.read_topic}"
        return result

    # Creates a new {SerialStream} using the parameters passed in the constructor
    def connect(self):
        self.stream = MqttStream(self.hostname, self.port, self.ssl, self.write_topic, self.read_topic, self.ack_timeout)
        if self.username:
            self.stream.username = self.username
        if self.password:
            self.stream.password = self.password
        if self.cert:
            self.stream.cert = self.cert
        if self.key:
            self.stream.key = self.key
        if self.ca_file:
            self.stream.ca_file = self.ca_file
        if self.keyfile_password:
            self.stream.keyfile_password = self.keyfile_password
        super().connect()

    # Supported Options
    # USERNAME - Username for Mqtt Server
    # PASSWORD - Password for Mqtt Server
    # CERT - Public Key for Client Cert Auth
    # KEY - Private Key for Client Cert Auth
    # CA_FILE - Certificate Authority for Client Cert Auth
    # KEYFILE_PASSWORD - Password for encrypted cert and key files
    # (see Interface#set_option)
    def set_option(self, option_name, option_values):
        super().set_option(option_name, option_values)
        match option_name.upper():
            case "ACK_TIMEOUT":
                self.ack_timeout = float(option_values[0])
            case "USERNAME":
                self.username = option_values[0]
            case "PASSWORD":
                self.password = option_values[0]
            case "CERT":
                # CERT must be given as a file
                self.cert = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.cert.write(option_values[0])
                self.cert.close()
            case "KEY":
                # KEY must be given as a file
                self.key = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.key.write(option_values[0])
                self.key.close()
            case "CA_FILE":
                # CA_FILE must be given as a file
                self.ca_file = tempfile.NamedTemporaryFile(mode="w+", delete=False)
                self.ca_file.write(option_values[0])
                self.ca_file.close()
            case "KEYFILE_PASSWORD":
                self.keyfile_password = option_values[0]



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.conversions.conversion import Conversion


class IpWriteConversion(Conversion):
    def __init__(self):
        super().__init__()
        self.converted_type = "UINT"
        self.converted_bit_size = 32

    # Perform the conversion on the value.
    #
    # @param value [Object] The value to convert
    # @param packet [Packet] The packet which contains the value. This can
    #   be useful to reach into the packet and use other values in the
    #   conversion.
    # @param buffer [String] The packet buffer
    # @return The converted value
    def call(self, value, _packet, _buffer):
        octets = value.split('.')
        return (int(octets[0]) << 24) | (int(octets[1]) << 16) | (int(octets[2]) << 8) | int(octets[3])

    # @return [String] The conversion class
    def __str__(self):
        return "IpWriteConversion"

    # @param read_or_write [String] Either 'READ' or 'WRITE'
    # @return [String] Config fragment for this conversion
    def to_config(read_or_write):
        return f"{read_or_write}_CONVERSION openc3/conversions/ip_write_conversion.py\n"



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from datetime import datetime, timezone
from openc3.models.model import Model

class ScriptStatusModel(Model):
    # Note: ScriptRunner only has permissions for keys that start with running-script
    RUNNING_PRIMARY_KEY = 'running-script'
    COMPLETED_PRIMARY_KEY = 'running-script-completed'

    @property
    def id(self):
        return self.name

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name, scope, type = "auto"):
        if type == "auto" or type == "running":
            # Check for running first
            running = super().get(f"{cls.RUNNING_PRIMARY_KEY}__{scope}", name = name)
            if running:
                return running
        return super().get(f"{cls.COMPLETED_PRIMARY_KEY}__{scope}", name = name)

    @classmethod
    def names(cls, scope, type = "running"):
        if type == "running":
            return super().names(f"{cls.RUNNING_PRIMARY_KEY}__{scope}")
        else:
            return super().names(f"{cls.COMPLETED_PRIMARY_KEY}__{scope}")

    @classmethod
    def all(cls, scope, offset = 0, limit = 10, type = "running"):
        if type == "running":
            keys = cls.store().zrevrange(f"{cls.RUNNING_PRIMARY_KEY}__{scope}__LIST", int(offset), int(offset) + int(limit) - 1)
            if len(keys) == 0:
                return []
            with cls.store().instance().redis_pool.get() as redis:
                pipeline = redis.pipeline(transaction=False)
                for key in keys:
                    pipeline.hget(f"{cls.RUNNING_PRIMARY_KEY}__{scope}", key)
                result = pipeline.execute()
                for i in range(len(result)):
                    if result[i] is not None:
                        result[i] = json.loads(result[i])
                return result
        else:
            keys = cls.store().zrevrange(f"{cls.COMPLETED_PRIMARY_KEY}__{scope}__LIST", int(offset), int(offset) + int(limit) - 1)
            if len(keys) == 0:
                return []
            with cls.store().instance().redis_pool.get() as redis:
                pipeline = redis.pipeline(transaction=False)
                for key in keys:
                    pipeline.hget(f"{cls.COMPLETED_PRIMARY_KEY}__{scope}", key)
                result = pipeline.execute()
                for i in range(len(result)):
                    if result[i] is not None:
                        result[i] = json.loads(result[i])
                return result

    @classmethod
    def count(cls, scope, type = "running"):
        if type == "running":
            return cls.store().zcount(f"{cls.RUNNING_PRIMARY_KEY}__#{scope}__LIST", 0, "+inf")
        else:
            return cls.store().zcount(f"{cls.COMPLETED_PRIMARY_KEY}__#{scope}__LIST", 0, "+inf")

    def __init__(
        self,
        name, # id
        state, # spawning, init, running, paused, waiting, error, breakpoint, crashed, stopped, completed, completed_errors, killed
        shard = 0, # Future enhancement of script runner shards
        filename = "", # The initial filename
        current_filename = None, # The current filename
        line_no = 0, # The current line number
        start_line_no = 1, # The line number to start the script at
        end_line_no = None, # The line number to end the script at
        username = None, # The username of the person who started the script
        user_full_name = None, # The full name of the person who started the script
        start_time = None, # The time the script started ISO format
        end_time = None, # The time the script ended ISO format
        disconnect = False,
        environment = None,
        suite_runner = None,
        errors = None,
        pid = None,
        log = None,
        report = None,
        updated_at = None,
        scope = None
    ):
        self.__state = state
        if self.is_complete():
            super().__init__(f"{self.COMPLETED_PRIMARY_KEY}__{scope}", name = name, updated_at = updated_at, plugin = None, scope = scope)
        else:
            super().__init__(f"{self.RUNNING_PRIMARY_KEY}__{scope}", name = name, updated_at = updated_at, plugin = None, scope = scope)
        self.shard = int(shard)
        self.filename = filename
        self.current_filename = current_filename
        self.line_no = line_no
        self.start_line_no = start_line_no
        self.end_line_no = end_line_no
        self.username = username
        self.user_full_name = user_full_name
        self.start_time = start_time
        self.end_time = end_time
        self.disconnect = disconnect
        self.environment = environment
        self.suite_runner = suite_runner
        self.errors = errors
        self.pid = pid
        self.log = log
        self.report = report

    def is_complete(self):
        return (self.__state == 'completed' or self.__state == 'completed_errors' or self.__state == 'stopped' or self.__state == 'crashed' or self.__state == 'killed')

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, new_state):
        # If the state is already a flavor of complete, leave it alone (first wins)
        if not self.is_complete():
            self.__state = new_state
            # If setting to complete, check for errors
            # and set the state to completed_errors if they exist
            if self.__state == 'completed' and self.errors:
                self.__state = 'completed_errors'

    # Update the Redis hash at primary_key and set the field "name"
    # to the JSON generated via calling as_json
    def create(self, update = False, force = False, queued = False, isoformat = True):
        self.updated_at = datetime.now(timezone.utc).isoformat()

        if queued:
            write_store = self.store_queued()
        else:
            write_store = self.store()
        write_store.hset(self.primary_key, self.name, json.dumps(self.as_json()))

        # Also add to ordered set on create
        if not update:
            mapping = {}
            mapping[self.name] = int(self.name)
            write_store.zadd(self.primary_key + "__LIST", mapping)

    def update(self, force = False, queued = False):
        # Magically handle the change from running to completed
        if self.is_complete() and self.primary_key == f"{self.RUNNING_PRIMARY_KEY}__{self.scope}":
            # Destroy the running key
            self.destroy(queued = queued)
            self.destroyed = False

            # Move to completed
            self.primary_key = f"{self.COMPLETED_PRIMARY_KEY}__{self.scope}"
            self.create(update = False, force = force, queued = queued, isoformat = True)
        else:
            self.create(update = True, force = force, queued = queued, isoformat = True)

    # Delete the model from the Store
    def destroy(self, queued = False):
        self.destroyed = True
        self.undeploy()
        if queued:
            write_store = self.store_queued()
        else:
            write_store = self.store()
        write_store.hdel(self.primary_key, self.name)
        # Also remove from ordered set
        write_store.zremrangebyscore(self.primary_key + "__LIST", int(self.name), int(self.name))

    def as_json(self):
        return {
            'name': self.name,
            'state': self.__state,
            'shard': self.shard,
            'filename': self.filename,
            'current_filename': self.current_filename,
            'line_no': self.line_no,
            'start_line_no': self.start_line_no,
            'end_line_no': self.end_line_no,
            'username': self.username,
            'user_full_name': self.user_full_name,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'disconnect': self.disconnect,
            'environment': self.environment,
            'suite_runner': self.suite_runner,
            'errors': self.errors,
            'pid': self.pid,
            'log': self.log,
            'report': self.report,
            'updated_at': self.updated_at,
            'scope': self.scope
        }
